"""Tests for the station registry module."""

import pytest

import weathervault as wv
from weathervault._registry import _sanitize_name, _StationNamespace, station


class TestSanitizeName:
    """Tests for the _sanitize_name function."""

    def test_basic_name(self):
        """Test simple station name."""
        assert _sanitize_name("LAGUARDIA") == "LAGUARDIA"

    def test_name_with_spaces(self):
        """Test name with spaces becomes underscores."""
        assert _sanitize_name("LA GUARDIA AIRPORT") == "LA_GUARDIA_AIRPORT"

    def test_name_with_special_chars(self):
        """Test special characters are replaced."""
        assert _sanitize_name("ST. LOUIS") == "ST_LOUIS"
        assert _sanitize_name("JOHN F. KENNEDY") == "JOHN_F_KENNEDY"

    def test_name_starting_with_number(self):
        """Test names starting with numbers get underscore prefix."""
        assert _sanitize_name("1ST STREET") == "_1ST_STREET"

    def test_empty_name(self):
        """Test empty name returns UNKNOWN."""
        assert _sanitize_name("") == "UNKNOWN"
        assert _sanitize_name(None) == "UNKNOWN"

    def test_lowercase_converted(self):
        """Test lowercase is converted to uppercase."""
        assert _sanitize_name("new york") == "NEW_YORK"


class TestStationNamespace:
    """Tests for the _StationNamespace class."""

    def test_access_station(self):
        """Test accessing a station returns its ID."""
        ns = _StationNamespace(stations={"TEST_STATION": "123456-99999"})
        assert ns.TEST_STATION == "123456-99999"

    def test_access_child_namespace(self):
        """Test accessing a child namespace."""
        child = _StationNamespace(stations={"CHILD_STATION": "111111-99999"})
        ns = _StationNamespace(children={"CHILD": child})
        assert child == ns.CHILD
        assert ns.CHILD.CHILD_STATION == "111111-99999"

    def test_dir_returns_available_names(self):
        """Test dir() returns station and child names."""
        child = _StationNamespace()
        ns = _StationNamespace(
            stations={"STATION_A": "1", "STATION_B": "2"},
            children={"GROUP_X": child},
        )
        names = dir(ns)
        assert "STATION_A" in names
        assert "STATION_B" in names
        assert "GROUP_X" in names

    def test_attribute_error_for_unknown(self):
        """Test AttributeError for unknown station."""
        ns = _StationNamespace(stations={"KNOWN": "123"})
        with pytest.raises(AttributeError):
            _ = ns.UNKNOWN

    def test_case_insensitive_access(self):
        """Test case-insensitive attribute access."""
        ns = _StationNamespace(stations={"TEST_STATION": "123456-99999"})
        assert ns.test_station == "123456-99999"

    def test_case_insensitive_child_access(self):
        """Test case-insensitive access for child namespaces."""
        child = _StationNamespace(stations={"INNER": "111"})
        ns = _StationNamespace(children={"CHILD_NS": child})
        # Access with different case
        assert ns.child_ns is child

    def test_case_insensitive_station_when_no_child_match(self):
        """Test case-insensitive station access when no child matches."""
        # Namespace has both children and stations with different names
        child = _StationNamespace(stations={"INNER": "111"})
        ns = _StationNamespace(
            children={"CHILD_NS": child},
            stations={"MY_STATION": "123456-99999"},
        )
        # Access station with different case (no child has this name)
        assert ns.my_station == "123456-99999"

    def test_private_attribute_raises(self):
        """Test that accessing private attributes raises AttributeError."""
        ns = _StationNamespace(stations={"TEST": "123"})
        with pytest.raises(AttributeError):
            _ = ns._private

    def test_repr_with_groups_only(self):
        """Test repr when namespace has only groups (children)."""
        child = _StationNamespace()
        ns = _StationNamespace(children={"GROUP_A": child, "GROUP_B": child})
        repr_str = repr(ns)
        assert "2 groups" in repr_str
        assert "stations" not in repr_str

    def test_repr_with_stations_only(self):
        """Test repr when namespace has only stations."""
        ns = _StationNamespace(stations={"STATION_A": "1", "STATION_B": "2", "STATION_C": "3"})
        repr_str = repr(ns)
        assert "3 stations" in repr_str
        assert "groups" not in repr_str

    def test_repr_with_both(self):
        """Test repr when namespace has both groups and stations."""
        child = _StationNamespace()
        ns = _StationNamespace(
            stations={"STATION_A": "1"},
            children={"GROUP_A": child},
        )
        repr_str = repr(ns)
        assert "1 groups" in repr_str
        assert "1 stations" in repr_str

    def test_repr_empty(self):
        """Test repr when namespace is empty."""
        ns = _StationNamespace()
        assert "empty" in repr(ns)

    def test_ipython_key_completions(self):
        """Test IPython bracket completion support."""
        ns = _StationNamespace(stations={"A": "1", "B": "2"})
        completions = ns._ipython_key_completions_()
        assert "A" in completions
        assert "B" in completions


class TestStationRegistry:
    """Tests for the station registry singleton."""

    def test_registry_exists(self):
        """Test that station registry is exported."""
        assert hasattr(wv, "station")
        assert wv.station is station

    def test_registry_repr_before_load(self):
        """Test repr shows not loaded initially."""
        # Create a fresh registry for this test
        from weathervault._registry import _StationRegistry

        fresh = _StationRegistry()
        assert "not loaded" in repr(fresh)

    @pytest.mark.network
    def test_access_us_namespace(self):
        """Test accessing US namespace loads data."""
        us = wv.station.US
        assert isinstance(us, _StationNamespace)

    @pytest.mark.network
    def test_access_us_state(self):
        """Test accessing a US state."""
        ny = wv.station.US.NY
        assert isinstance(ny, _StationNamespace)
        # Should have stations
        assert len(dir(ny)) > 0

    @pytest.mark.network
    def test_registry_dir_after_load(self):
        """Test dir() on registry after loading."""
        # Access to trigger load
        _ = wv.station.US
        countries = dir(wv.station)
        assert "US" in countries
        assert len(countries) > 10  # Should have many countries

    @pytest.mark.network
    def test_registry_repr_after_load(self):
        """Test repr shows country count after loading."""
        # Access to trigger load
        _ = wv.station.US
        repr_str = repr(wv.station)
        assert "countries" in repr_str
        assert "not loaded" not in repr_str

    @pytest.mark.network
    def test_get_known_station_id(self):
        """Test getting a known station ID."""
        # JFK airport should be available
        jfk_id = wv.station.US.NY.JOHN_F_KENNEDY_INTERNATIONAL_AIRPORT
        assert jfk_id == "744860-94789"

    @pytest.mark.network
    def test_search_functionality(self):
        """Test the search method."""
        results = wv.station.search("kennedy")
        assert len(results) > 0
        # Should find JFK
        assert any("KENNEDY" in key for key in results)

    @pytest.mark.network
    def test_active_station_preferred(self):
        """Test that active stations are preferred over historical ones."""
        # LA GUARDIA AIRPORT (725030-14732) should be preferred over
        # NEW YORK LAGUARDIA ARPT (999999-14732) which ended in 1972
        lga_id = wv.station.US.NY.LA_GUARDIA_AIRPORT
        assert lga_id == "725030-14732"

    @pytest.mark.network
    def test_non_us_country(self):
        """Test accessing non-US countries."""
        # Germany should have stations (ISO code DE, FIPS code GM)
        de = wv.station.DE  # Germany uses ISO code DE
        assert hasattr(de, "__getattr__")  # Should be a country node

    @pytest.mark.network
    def test_use_with_get_weather_data(self):
        """Test that station IDs work with get_weather_data."""
        station_id = wv.station.US.NY.JOHN_F_KENNEDY_INTERNATIONAL_AIRPORT
        # Just verify it's a valid string format
        assert isinstance(station_id, str)
        assert "-" in station_id
        parts = station_id.split("-")
        assert len(parts) == 2
        assert parts[0].isdigit()
        assert parts[1].isdigit()



#!/usr/bin/env python
"""Download sample data for documentation examples.

This script downloads weather data files that are bundled with the package
to enable documentation examples to run without network access.

Buffer years (±1) are included because when converting to local time,
data from adjacent years may be needed to ensure complete coverage at
year boundaries.

Usage:
    python scripts/download_sample_data.py
"""

from pathlib import Path

import httpx

BASE_URL = "https://www.ncei.noaa.gov/pub/data/noaa"

# Stations and years used in documentation examples
# Format: (station_id, [years]) - buffer years (±1) are automatically added
EXAMPLE_STATIONS = [
    ("744860-94789", [2023]),  # JFK Airport (get_weather_data basic example)
    ("725030-14732", [2022, 2023]),  # LaGuardia Airport (multi-year, cache, incl_stn_info)
    ("724050-13743", [2023]),  # Washington Reagan Airport (make_hourly example)
]


def _get_years_with_buffers(years: list[int]) -> list[int]:
    """Add buffer years (±1) to handle timezone conversion at year boundaries."""
    all_years = set(years)
    for year in years:
        all_years.add(year - 1)
        all_years.add(year + 1)
    return sorted(all_years)


def download_sample_data():
    """Download sample data files to the package data directory."""
    data_dir = Path(__file__).parent.parent / "weathervault" / "data"
    data_dir.mkdir(parents=True, exist_ok=True)

    # Build full list with buffer years
    sample_data = []
    for station_id, years in EXAMPLE_STATIONS:
        for year in _get_years_with_buffers(years):
            sample_data.append((station_id, year))

    print(f"Downloading sample data to: {data_dir}")
    print("(includes buffer years for timezone handling)")
    print()

    with httpx.Client(timeout=120.0, follow_redirects=True) as client:
        for station_id, year in sample_data:
            filename = f"{station_id}-{year}.gz"
            filepath = data_dir / filename
            url = f"{BASE_URL}/{year}/{filename}"

            if filepath.exists():
                print(f"  ✓ {filename} (already exists)")
                continue

            print(f"  ↓ Downloading {filename}...", end=" ", flush=True)
            try:
                response = client.get(url)
                response.raise_for_status()
                filepath.write_bytes(response.content)
                size_kb = len(response.content) / 1024
                print(f"({size_kb:.1f} KB)")
            except httpx.HTTPStatusError as e:
                if e.response.status_code == 404:
                    print("(not available - OK for buffer year)")
                else:
                    print(f"FAILED: {e}")
            except httpx.HTTPError as e:
                print(f"FAILED: {e}")

    print()
    print("Done! Sample data is ready for documentation builds.")


if __name__ == "__main__":
    download_sample_data()



"""Station metadata and inventory functions."""

from __future__ import annotations

import io
from typing import TYPE_CHECKING

import httpx
import polars as pl
from timezonefinder import TimezoneFinder

from weathervault._constants import BASE_URL, COUNTRY_CODES, FIPS_TO_ISO

if TYPE_CHECKING:
    pass

# Module-level cache for station data (always stores version WITH timezone)
_station_metadata_cache: pl.DataFrame | None = None
_station_metadata_cache_has_tz: bool = False
_inventory_cache: pl.DataFrame | None = None
_tf: TimezoneFinder | None = None


def _get_timezone_finder() -> TimezoneFinder:
    """Get or create the TimezoneFinder instance."""
    global _tf
    if _tf is None:
        _tf = TimezoneFinder()
    return _tf


def get_station_metadata(
    *,
    include_timezone: bool = True,
    force_refresh: bool = False,
) -> pl.DataFrame:
    """
    Get metadata for all weather stations.

    Downloads and parses the ISD history file containing information about all weather stations in
    the database.

    Parameters
    ----------
    include_timezone
        Whether to look up timezone names based on station coordinates. This adds a `tz_name`
        column. Default is `True`.
    force_refresh
        Force re-download of station data even if cached. Default is `False`.

    Returns
    -------
    pl.DataFrame
        A DataFrame with station metadata including:

        - `id`: Station ID (USAF-WBAN format)
        - `usaf`: USAF station identifier
        - `wban`: WBAN station identifier
        - `name`: Station name
        - `country_code`: ISO 3166-1 alpha-2 country code
        - `country`: Country name
        - `state`: US state code (if applicable)
        - `icao`: ICAO identifier
        - `lat`: Latitude in degrees
        - `lon`: Longitude in degrees
        - `elev`: Elevation in meters
        - `begin_date`: Date of first available data
        - `end_date`: Date of last available data
        - `tz_name`: Timezone name (if `include_timezone=` is `True`)

    Examples
    --------
    Get metadata for all weather stations:

    ```{python}
    import weathervault as wv
    stations = wv.get_station_metadata()
    stations.head()
    ```
    """
    global _station_metadata_cache, _station_metadata_cache_has_tz

    # If we have a cached version with timezone, we can serve both cases
    if _station_metadata_cache is not None and not force_refresh:
        if _station_metadata_cache_has_tz:
            # Cache has timezone: can serve both requests
            if include_timezone:
                return _station_metadata_cache
            else:
                # Return without tz_name column
                return _station_metadata_cache.drop("tz_name")
        elif not include_timezone:
            # Cache doesn't have timezone, and we don't need it
            return _station_metadata_cache
        # Otherwise, cache doesn't have timezone but we need it: refetch

    url = f"{BASE_URL}/isd-history.csv"

    with httpx.Client(timeout=60.0) as client:
        response = client.get(url)
        response.raise_for_status()

    df = pl.read_csv(
        io.BytesIO(response.content),
        schema_overrides={
            "USAF": pl.Utf8,
            "WBAN": pl.Utf8,
            "BEGIN": pl.Utf8,
            "END": pl.Utf8,
        },
    )

    # Rename columns to be more Pythonic
    df = df.rename(
        {
            "USAF": "usaf",
            "WBAN": "wban",
            "STATION NAME": "name",
            "CTRY": "country_code",
            "STATE": "state",
            "ICAO": "icao",
            "LAT": "lat",
            "LON": "lon",
            "ELEV(M)": "elev",
            "BEGIN": "begin_date",
            "END": "end_date",
        }
    )

    # Create station ID and parse dates
    df = df.with_columns(
        pl.concat_str([pl.col("usaf"), pl.lit("-"), pl.col("wban")]).alias("id"),
        pl.col("begin_date").str.to_date("%Y%m%d").alias("begin_date"),
        pl.col("end_date").str.to_date("%Y%m%d").alias("end_date"),
        # Ensure lat/lon/elev are numeric (they may have + prefix in CSV)
        pl.col("lat").cast(pl.Float64, strict=False),
        pl.col("lon").cast(pl.Float64, strict=False),
        pl.col("elev").cast(pl.Float64, strict=False),
    )

    # Replace missing value sentinel (-999.0, -999.9) with null for elevation
    df = df.with_columns(
        pl.when(pl.col("elev") < -900).then(None).otherwise(pl.col("elev")).alias("elev")
    )

    # Add country name from FIPS code (before conversion to ISO)
    df = df.with_columns(
        pl.col("country_code").replace_strict(COUNTRY_CODES, default=None).alias("country")
    )

    # Convert FIPS country codes to ISO codes
    df = df.with_columns(
        pl.col("country_code").replace_strict(FIPS_TO_ISO, default=None).alias("country_code")
    )

    # Reorder columns
    column_order = [
        "id",
        "usaf",
        "wban",
        "name",
        "country_code",
        "country",
        "state",
        "icao",
        "lat",
        "lon",
        "elev",
        "begin_date",
        "end_date",
    ]

    if include_timezone:
        # Look up timezone for each station based on coordinates
        tf = _get_timezone_finder()

        def get_tz(lat: float | None, lon: float | None) -> str | None:
            if lat is None or lon is None:
                return None
            try:
                return tf.timezone_at(lat=lat, lng=lon)
            except Exception:  # pragma: no cover
                return None

        tz_names = [
            get_tz(lat, lon)
            for lat, lon in zip(df["lat"].to_list(), df["lon"].to_list(), strict=True)
        ]
        df = df.with_columns(pl.Series("tz_name", tz_names))
        column_order.append("tz_name")

    df = df.select(column_order)

    # Cache the result and track whether it has timezone
    _station_metadata_cache = df
    _station_metadata_cache_has_tz = include_timezone
    return df


def get_inventory(*, force_refresh: bool = False) -> pl.DataFrame:
    """
    Get the data inventory for all stations and years.

    Downloads and parses the ISD inventory file containing record counts for each station by year
    and month.

    Parameters
    ----------
    force_refresh
        Force re-download of inventory data even if cached. Default is `False`.

    Returns
    -------
    pl.DataFrame
        A DataFrame with inventory information including:

        - `id`: Station ID (USAF-WBAN format)
        - `usaf`: USAF station identifier
        - `wban`: WBAN station identifier
        - `year`: Year of data
        - `jan` through `dec`: Record counts for each month
        - `total`: Total records for the year

    Examples
    --------
    Get data inventory for all stations:

    ```{python}
    import weathervault as wv
    inventory = wv.get_inventory()
    inventory.filter(pl.col("id") == "725030-14732")
    ```
    """
    global _inventory_cache

    if _inventory_cache is not None and not force_refresh:
        return _inventory_cache

    url = f"{BASE_URL}/isd-inventory.csv"

    with httpx.Client(timeout=60.0) as client:
        response = client.get(url)
        response.raise_for_status()

    df = pl.read_csv(
        io.BytesIO(response.content),
        schema_overrides={
            "USAF": pl.Utf8,
            "WBAN": pl.Utf8,
        },
    )

    # Rename columns
    df = df.rename(
        {
            "USAF": "usaf",
            "WBAN": "wban",
            "YEAR": "year",
            "JAN": "jan",
            "FEB": "feb",
            "MAR": "mar",
            "APR": "apr",
            "MAY": "may",
            "JUN": "jun",
            "JUL": "jul",
            "AUG": "aug",
            "SEP": "sep",
            "OCT": "oct",
            "NOV": "nov",
            "DEC": "dec",
        }
    )

    # Create station ID and calculate total
    df = df.with_columns(
        pl.concat_str([pl.col("usaf"), pl.lit("-"), pl.col("wban")]).alias("id"),
        (
            pl.col("jan")
            + pl.col("feb")
            + pl.col("mar")
            + pl.col("apr")
            + pl.col("may")
            + pl.col("jun")
            + pl.col("jul")
            + pl.col("aug")
            + pl.col("sep")
            + pl.col("oct")
            + pl.col("nov")
            + pl.col("dec")
        ).alias("total"),
    )

    # Reorder columns
    df = df.select(
        [
            "id",
            "usaf",
            "wban",
            "year",
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
            "total",
        ]
    )

    _inventory_cache = df
    return df


def search_stations(
    *,
    name: str | None = None,
    country: str | None = None,
    country_code: str | None = None,
    state: str | None = None,
    lat_range: tuple[float, float] | None = None,
    lon_range: tuple[float, float] | None = None,
    has_recent_data: bool = False,
) -> pl.DataFrame:
    """
    Search for weather stations matching specified criteria.

    Parameters
    ----------
    name
        Search for stations with names containing this string (case-insensitive).
    country
        Filter by country name (case-insensitive, partial match).
    country_code
        Filter by ISO 3166-1 alpha-2 country code (e.g., `"US"`, `"GB"`, `"FR"`).
        For convenience, you can also use the `wv.country` shorthand:
        `wv.country.US`, `wv.country.GB`, `wv.country.DE`, etc.
    state
        Filter by US state or territory code (exact match, e.g., `"CA"`, `"NY"`, `"TX"`).
        For convenience, you can also use the `wv.state` shorthand:
        `wv.state.CA`, `wv.state.NY`, `wv.state.TX`, etc.
    lat_range
        Filter by latitude range as (min_lat, max_lat).
    lon_range
        Filter by longitude range as (min_lon, max_lon).
    has_recent_data
        If `True`, only return stations with data in the current or previous year.

    Returns
    -------
    pl.DataFrame
        A DataFrame of stations matching the search criteria.

    Examples
    --------
    Search for stations in California using the state shorthand:

    ```{python}
    import weathervault as wv

    # Using wv.state shorthand (provides autocomplete)
    ca_stations = wv.search_stations(state=wv.state.CA)

    ca_stations
    ```

    Search by country code using the convenient shorthand:

    ```{python}
    # Using wv.country shorthand (provides autocomplete)
    de_stations = wv.search_stations(country_code=wv.country.DE)

    de_stations
    ```

    Combine state and country filters:

    ```{python}
    # Find stations in California, US
    us_ca = wv.search_stations(
        country_code=wv.country.US,
        state=wv.state.CA
    )

    us_ca
    ```

    Search for stations within a bounding box:

    ```{python}
    within = wv.search_stations(
        lat_range=(37.0, 38.0),
        lon_range=(-122.5, -121.5)
    )

    within
    ```
    """
    df = get_station_metadata()

    if name is not None:
        df = df.filter(pl.col("name").str.to_lowercase().str.contains(name.lower()))

    if country is not None:
        df = df.filter(pl.col("country").str.to_lowercase().str.contains(country.lower()))

    if country_code is not None:
        # Filter by ISO country code (DataFrame now contains ISO codes)
        iso_code = country_code.upper()
        df = df.filter(pl.col("country_code") == iso_code)

    if state is not None:
        df = df.filter(pl.col("state") == state.upper())

    if lat_range is not None:
        min_lat, max_lat = lat_range
        df = df.filter((pl.col("lat") >= min_lat) & (pl.col("lat") <= max_lat))

    if lon_range is not None:
        min_lon, max_lon = lon_range
        df = df.filter((pl.col("lon") >= min_lon) & (pl.col("lon") <= max_lon))

    if has_recent_data:
        from datetime import date

        current_year = date.today().year
        cutoff_date = date(current_year - 1, 1, 1)
        df = df.filter(pl.col("end_date") >= cutoff_date)

    return df


def get_years_for_station(station_id: str) -> list[int]:
    """
    Get the list of years for which a station has data.

    Parameters
    ----------
    station_id
        The station ID in USAF-WBAN format (e.g., `"725030-14732"`).

    Returns
    -------
    list[int]
        List of years with available data.

    Examples
    --------
    Get available years for a specific station:

    ```{python}
    import weathervault as wv

    years = wv.get_years_for_station("725030-14732")

    years
    ```
    """
    inventory = get_inventory()
    station_data = inventory.filter(pl.col("id") == station_id)

    if station_data.height == 0:
        return []

    return sorted(station_data["year"].to_list())


def get_countries() -> pl.DataFrame:
    """
    Get a DataFrame of country codes and names.

    Returns
    -------
    pl.DataFrame
        A DataFrame with columns:

        - country_code: ISO 3166-1 alpha-2 country code
        - country: Country name

    Examples
    --------
    Get a list of all countries with weather stations:

    ```{python}
    import weathervault as wv

    countries = wv.get_countries()

    countries
    ```
    """
    # Create mapping from FIPS to ISO codes with country names
    iso_codes = []
    country_names = []

    for fips_code, country_name in COUNTRY_CODES.items():
        if fips_code in FIPS_TO_ISO:  # pragma: no branch
            iso_codes.append(FIPS_TO_ISO[fips_code])
            country_names.append(country_name)

    return pl.DataFrame(
        {
            "country_code": iso_codes,
            "country": country_names,
        }
    ).sort("country")



"""Quick test that bundled data works offline."""

import weathervault as wv

# Test: This should use bundled data (no network for the data itself!)
# Note: Station metadata still requires network on first call
print("Testing bundled data access...")

# Check what bundled years exist
from weathervault.weather import _get_bundled_years

for station in ["744860-94789", "725030-14732", "724050-13743"]:
    years = _get_bundled_years(station)
    print(f"  {station}: bundled years = {sorted(years)}")

print()
print("Fetching LaGuardia 2023 data...")
weather = wv.get_weather_data("725030-14732", years=2023)
print(f"Got {weather.height} rows")
print(weather.head(3))



"""Constants for weathervault package."""

# Base URL for NCEI NOAA data
BASE_URL = "https://www.ncei.noaa.gov/pub/data/noaa"

# Column widths for the fixed-width mandatory data section
# Based on ISD format document
MANDATORY_COLUMN_WIDTHS = [
    4,  # Variable length indicator (total variable length)
    6,  # USAF station ID
    5,  # WBAN station ID
    4,  # Year
    2,  # Month
    2,  # Day
    2,  # Hour
    2,  # Minute
    1,  # Data source flag
    6,  # Latitude (scaled)
    7,  # Longitude (scaled)
    5,  # Report type code
    5,  # Elevation (scaled)
    5,  # Call letter identifier
    4,  # Quality control process name
    3,  # Wind direction angle
    1,  # Wind direction quality code
    1,  # Wind type code
    4,  # Wind speed rate (scaled)
    1,  # Wind speed quality code
    5,  # Ceiling height (scaled)
    1,  # Ceiling height quality code
    1,  # Ceiling determination code
    1,  # CAVOK code
    6,  # Visibility distance (scaled)
    1,  # Visibility distance quality code
    1,  # Visibility variability code
    1,  # Visibility variability quality code
    5,  # Air temperature (scaled)
    1,  # Air temperature quality code
    5,  # Dew point temperature (scaled)
    1,  # Dew point quality code
    5,  # Sea level pressure (scaled)
    1,  # Sea level pressure quality code
]

# Column names for the mandatory section
MANDATORY_COLUMN_NAMES = [
    "total_chars",
    "usaf",
    "wban",
    "year",
    "month",
    "day",
    "hour",
    "minute",
    "data_source",
    "latitude",
    "longitude",
    "report_type",
    "elevation",
    "call_letters",
    "qc_process",
    "wind_direction",
    "wind_direction_qc",
    "wind_type",
    "wind_speed",
    "wind_speed_qc",
    "ceiling_height",
    "ceiling_height_qc",
    "ceiling_determination",
    "cavok",
    "visibility",
    "visibility_qc",
    "visibility_variability",
    "visibility_variability_qc",
    "temp",
    "temp_qc",
    "dew_point",
    "dew_point_qc",
    "sea_level_pressure",
    "sea_level_pressure_qc",
]

# Missing value codes in ISD data
# Reference: NOAA ISD Format Document
# https://www.ncei.noaa.gov/data/global-hourly/doc/isd-format-document.pdf
MISSING_VALUES = {
    "wind_direction": 999,  # POS 61-63: 999 = Missing
    "wind_speed": 9999,  # POS 66-69: 9999 = Missing
    "ceiling_height": 99999,  # POS 71-75: 99999 = Missing
    "visibility": 999999,  # POS 79-84: 999999 = Missing
    "temp": 9999,  # POS 88-92: +9999 = Missing (stored without sign)
    "dew_point": 9999,  # POS 94-98: +9999 = Missing (stored without sign)
    "sea_level_pressure": 99999,  # POS 100-104: 99999 = Missing
}

# Scale factors for numeric fields (divide by this to get actual value)
SCALE_FACTORS = {
    "wind_speed": 10,  # m/s with 1 decimal
    "temp": 10,  # degrees C with 1 decimal
    "dew_point": 10,  # degrees C with 1 decimal
    "sea_level_pressure": 10,  # hectopascals with 1 decimal
    "latitude": 1000,  # degrees with 3 decimals
    "longitude": 1000,  # degrees with 3 decimals
    "elevation": 1,  # meters (already in meters)
}

# FIPS country codes to country names mapping
# From https://www.ncei.noaa.gov/pub/data/noaa/country-list.txt
COUNTRY_CODES = {
    "AA": "Aruba",
    "AC": "Antigua and Barbuda",
    "AF": "Afghanistan",
    "AG": "Algeria",
    "AI": "Ascension Island",
    "AJ": "Azerbaijan",
    "AL": "Albania",
    "AM": "Armenia",
    "AN": "Andorra",
    "AO": "Angola",
    "AQ": "American Samoa",
    "AR": "Argentina",
    "AS": "Australia",
    "AT": "Ashmore and Cartier Islands",
    "AU": "Austria",
    "AV": "Anguilla",
    "AY": "Antarctica",
    "AZ": "Azores",
    "BA": "Bahrain",
    "BB": "Barbados",
    "BC": "Botswana",
    "BD": "Bermuda",
    "BE": "Belgium",
    "BF": "Bahamas",
    "BG": "Bangladesh",
    "BH": "Belize",
    "BK": "Bosnia and Herzegovina",
    "BL": "Bolivia",
    "BM": "Burma",
    "BN": "Benin",
    "BO": "Belarus",
    "BP": "Solomon Islands",
    "BR": "Brazil",
    "BT": "Bhutan",
    "BU": "Bulgaria",
    "BV": "Bouvet Island",
    "BX": "Brunei",
    "BY": "Burundi",
    "CA": "Canada",
    "CB": "Cambodia",
    "CD": "Chad",
    "CE": "Sri Lanka",
    "CF": "Congo",
    "CG": "Zaire",
    "CH": "China",
    "CI": "Chile",
    "CJ": "Cayman Islands",
    "CK": "Cocos (Keeling) Islands",
    "CM": "Cameroon",
    "CN": "Comoros",
    "CO": "Colombia",
    "CQ": "Northern Mariana Islands",
    "CR": "Coral Sea Islands",
    "CS": "Costa Rica",
    "CT": "Central African Republic",
    "CU": "Cuba",
    "CV": "Cape Verde",
    "CW": "Cook Islands",
    "CY": "Cyprus",
    "DA": "Denmark",
    "DJ": "Djibouti",
    "DO": "Dominica",
    "DR": "Dominican Republic",
    "EC": "Ecuador",
    "EG": "Egypt",
    "EI": "Ireland",
    "EK": "Equatorial Guinea",
    "EN": "Estonia",
    "ER": "Eritrea",
    "ES": "El Salvador",
    "ET": "Ethiopia",
    "EZ": "Czech Republic",
    "FG": "French Guiana",
    "FI": "Finland",
    "FJ": "Fiji",
    "FK": "Falkland Islands",
    "FM": "Micronesia",
    "FO": "Faroe Islands",
    "FP": "French Polynesia",
    "FR": "France",
    "GA": "Gambia",
    "GB": "Gabon",
    "GG": "Georgia",
    "GH": "Ghana",
    "GI": "Gibraltar",
    "GJ": "Grenada",
    "GK": "Guernsey",
    "GL": "Greenland",
    "GM": "Germany",
    "GP": "Guadeloupe",
    "GQ": "Guam",
    "GR": "Greece",
    "GT": "Guatemala",
    "GV": "Guinea",
    "GY": "Guyana",
    "GZ": "Gaza Strip",
    "HA": "Haiti",
    "HK": "Hong Kong",
    "HO": "Honduras",
    "HR": "Croatia",
    "HU": "Hungary",
    "IC": "Iceland",
    "ID": "Indonesia",
    "IM": "Isle of Man",
    "IN": "India",
    "IO": "British Indian Ocean Territory",
    "IR": "Iran",
    "IS": "Israel",
    "IT": "Italy",
    "IV": "Cote d'Ivoire",
    "IZ": "Iraq",
    "JA": "Japan",
    "JE": "Jersey",
    "JM": "Jamaica",
    "JN": "Jan Mayen",
    "JO": "Jordan",
    "KE": "Kenya",
    "KG": "Kyrgyzstan",
    "KN": "North Korea",
    "KR": "Kiribati",
    "KS": "South Korea",
    "KT": "Christmas Island",
    "KU": "Kuwait",
    "KV": "Kosovo",
    "KZ": "Kazakhstan",
    "LA": "Laos",
    "LE": "Lebanon",
    "LG": "Latvia",
    "LH": "Lithuania",
    "LI": "Liberia",
    "LO": "Slovakia",
    "LS": "Liechtenstein",
    "LT": "Lesotho",
    "LU": "Luxembourg",
    "LY": "Libya",
    "MA": "Madagascar",
    "MB": "Martinique",
    "MC": "Macau",
    "MD": "Moldova",
    "MF": "Mayotte",
    "MG": "Mongolia",
    "MH": "Montserrat",
    "MI": "Malawi",
    "MJ": "Montenegro",
    "MK": "North Macedonia",
    "ML": "Mali",
    "MM": "Burma (Myanmar)",
    "MN": "Monaco",
    "MO": "Morocco",
    "MP": "Mauritius",
    "MR": "Mauritania",
    "MT": "Malta",
    "MU": "Oman",
    "MV": "Maldives",
    "MX": "Mexico",
    "MY": "Malaysia",
    "MZ": "Mozambique",
    "NC": "New Caledonia",
    "NE": "Niue",
    "NF": "Norfolk Island",
    "NG": "Niger",
    "NH": "Vanuatu",
    "NI": "Nigeria",
    "NL": "Netherlands",
    "NO": "Norway",
    "NP": "Nepal",
    "NR": "Nauru",
    "NS": "Suriname",
    "NU": "Nicaragua",
    "NZ": "New Zealand",
    "OD": "South Sudan",
    "PA": "Paraguay",
    "PC": "Pitcairn Islands",
    "PE": "Peru",
    "PG": "Spratly Islands",
    "PK": "Pakistan",
    "PL": "Poland",
    "PM": "Panama",
    "PO": "Portugal",
    "PP": "Papua New Guinea",
    "PS": "Palau",
    "PU": "Guinea-Bissau",
    "QA": "Qatar",
    "RE": "Reunion",
    "RI": "Serbia",
    "RM": "Marshall Islands",
    "RO": "Romania",
    "RP": "Philippines",
    "RQ": "Puerto Rico",
    "RS": "Russia",
    "RW": "Rwanda",
    "SA": "Saudi Arabia",
    "SB": "St. Pierre and Miquelon",
    "SC": "St. Kitts and Nevis",
    "SE": "Seychelles",
    "SF": "South Africa",
    "SG": "Senegal",
    "SH": "St. Helena",
    "SI": "Slovenia",
    "SL": "Sierra Leone",
    "SM": "San Marino",
    "SN": "Singapore",
    "SO": "Somalia",
    "SP": "Spain",
    "SR": "Serbia",
    "ST": "St. Lucia",
    "SU": "Sudan",
    "SV": "Svalbard",
    "SW": "Sweden",
    "SX": "South Georgia",
    "SY": "Syria",
    "SZ": "Switzerland",
    "TC": "United Arab Emirates",
    "TD": "Trinidad and Tobago",
    "TH": "Thailand",
    "TI": "Tajikistan",
    "TK": "Turks and Caicos Islands",
    "TL": "Tokelau",
    "TN": "Tonga",
    "TO": "Togo",
    "TP": "Sao Tome and Principe",
    "TS": "Tunisia",
    "TU": "Turkey",
    "TV": "Tuvalu",
    "TW": "Taiwan",
    "TX": "Turkmenistan",
    "TZ": "Tanzania",
    "UG": "Uganda",
    "UK": "United Kingdom",
    "UP": "Ukraine",
    "US": "United States",
    "UV": "Burkina Faso",
    "UY": "Uruguay",
    "UZ": "Uzbekistan",
    "VC": "St. Vincent and the Grenadines",
    "VE": "Venezuela",
    "VI": "Virgin Islands (British)",
    "VM": "Vietnam",
    "VQ": "Virgin Islands (U.S.)",
    "VT": "Vatican City",
    "WA": "Namibia",
    "WE": "West Bank",
    "WF": "Wallis and Futuna",
    "WI": "Western Sahara",
    "WQ": "Wake Island",
    "WS": "Western Samoa",
    "WZ": "Eswatini",
    "YM": "Yemen",
    "ZA": "Zambia",
    "ZI": "Zimbabwe",
    "ZM": "Samoa",
}

# FIPS to ISO 3166-1 alpha-2 country code mapping
# Maps the FIPS codes used in ISD data to standard ISO country codes
# This allows users to access stations using familiar ISO codes (e.g., DE for Germany)
# while internally working with the FIPS codes in the data
FIPS_TO_ISO = {
    "AA": "AW",  # Aruba
    "AC": "AG",  # Antigua and Barbuda
    "AF": "AF",  # Afghanistan (same)
    "AG": "DZ",  # Algeria
    "AI": "AC",  # Ascension Island (unofficial)
    "AJ": "AZ",  # Azerbaijan
    "AL": "AL",  # Albania (same)
    "AM": "AM",  # Armenia (same)
    "AN": "AD",  # Andorra
    "AO": "AO",  # Angola (same)
    "AQ": "AS",  # American Samoa
    "AR": "AR",  # Argentina (same)
    "AS": "AU",  # Australia
    "AT": "AU",  # Ashmore and Cartier Islands (part of Australia)
    "AU": "AT",  # Austria
    "AV": "AI",  # Anguilla
    "AY": "AQ",  # Antarctica
    "AZ": "PT",  # Azores (part of Portugal)
    "BA": "BH",  # Bahrain
    "BB": "BB",  # Barbados (same)
    "BC": "BW",  # Botswana
    "BD": "BM",  # Bermuda
    "BE": "BE",  # Belgium (same)
    "BF": "BS",  # Bahamas
    "BG": "BD",  # Bangladesh
    "BH": "BZ",  # Belize
    "BK": "BA",  # Bosnia and Herzegovina
    "BL": "BO",  # Bolivia
    "BM": "MM",  # Burma/Myanmar
    "BN": "BJ",  # Benin
    "BO": "BY",  # Belarus
    "BP": "SB",  # Solomon Islands
    "BR": "BR",  # Brazil (same)
    "BT": "BT",  # Bhutan (same)
    "BU": "BG",  # Bulgaria
    "BV": "BV",  # Bouvet Island (same)
    "BX": "BN",  # Brunei
    "BY": "BI",  # Burundi
    "CA": "CA",  # Canada (same)
    "CB": "KH",  # Cambodia
    "CD": "TD",  # Chad
    "CE": "LK",  # Sri Lanka
    "CF": "CG",  # Congo
    "CG": "CD",  # Zaire/DR Congo
    "CH": "CN",  # China
    "CI": "CL",  # Chile
    "CJ": "KY",  # Cayman Islands
    "CK": "CC",  # Cocos (Keeling) Islands
    "CM": "CM",  # Cameroon (same)
    "CN": "KM",  # Comoros
    "CO": "CO",  # Colombia (same)
    "CQ": "MP",  # Northern Mariana Islands
    "CR": "AU",  # Coral Sea Islands (part of Australia)
    "CS": "CR",  # Costa Rica
    "CT": "CF",  # Central African Republic
    "CU": "CU",  # Cuba (same)
    "CV": "CV",  # Cape Verde (same)
    "CW": "CK",  # Cook Islands
    "CY": "CY",  # Cyprus (same)
    "DA": "DK",  # Denmark
    "DJ": "DJ",  # Djibouti (same)
    "DO": "DM",  # Dominica
    "DR": "DO",  # Dominican Republic
    "EC": "EC",  # Ecuador (same)
    "EG": "EG",  # Egypt (same)
    "EI": "IE",  # Ireland
    "EK": "GQ",  # Equatorial Guinea
    "EN": "EE",  # Estonia
    "ER": "ER",  # Eritrea (same)
    "ES": "SV",  # El Salvador
    "ET": "ET",  # Ethiopia (same)
    "EZ": "CZ",  # Czech Republic
    "FG": "GF",  # French Guiana
    "FI": "FI",  # Finland (same)
    "FJ": "FJ",  # Fiji (same)
    "FK": "FK",  # Falkland Islands (same)
    "FM": "FM",  # Micronesia (same)
    "FO": "FO",  # Faroe Islands (same)
    "FP": "PF",  # French Polynesia
    "FR": "FR",  # France (same)
    "GA": "GM",  # Gambia
    "GB": "GA",  # Gabon
    "GG": "GE",  # Georgia
    "GH": "GH",  # Ghana (same)
    "GI": "GI",  # Gibraltar (same)
    "GJ": "GD",  # Grenada
    "GK": "GG",  # Guernsey
    "GL": "GL",  # Greenland (same)
    "GM": "DE",  # Germany
    "GP": "GP",  # Guadeloupe (same)
    "GQ": "GU",  # Guam
    "GR": "GR",  # Greece (same)
    "GT": "GT",  # Guatemala (same)
    "GV": "GN",  # Guinea
    "GY": "GY",  # Guyana (same)
    "GZ": "PS",  # Gaza Strip (Palestine)
    "HA": "HT",  # Haiti
    "HK": "HK",  # Hong Kong (same)
    "HO": "HN",  # Honduras
    "HR": "HR",  # Croatia (same)
    "HU": "HU",  # Hungary (same)
    "IC": "IS",  # Iceland
    "ID": "ID",  # Indonesia (same)
    "IM": "IM",  # Isle of Man (same)
    "IN": "IN",  # India (same)
    "IO": "IO",  # British Indian Ocean Territory (same)
    "IR": "IR",  # Iran (same)
    "IS": "IL",  # Israel
    "IT": "IT",  # Italy (same)
    "IV": "CI",  # Cote d'Ivoire
    "IZ": "IQ",  # Iraq
    "JA": "JP",  # Japan
    "JE": "JE",  # Jersey (same)
    "JM": "JM",  # Jamaica (same)
    "JN": "SJ",  # Jan Mayen (part of Svalbard and Jan Mayen)
    "JO": "JO",  # Jordan (same)
    "KE": "KE",  # Kenya (same)
    "KG": "KG",  # Kyrgyzstan (same)
    "KN": "KP",  # North Korea
    "KR": "KI",  # Kiribati
    "KS": "KR",  # South Korea
    "KT": "CX",  # Christmas Island
    "KU": "KW",  # Kuwait
    "KV": "XK",  # Kosovo (unofficial)
    "KZ": "KZ",  # Kazakhstan (same)
    "LA": "LA",  # Laos (same)
    "LE": "LB",  # Lebanon
    "LG": "LV",  # Latvia
    "LH": "LT",  # Lithuania
    "LI": "LR",  # Liberia
    "LO": "SK",  # Slovakia
    "LS": "LI",  # Liechtenstein
    "LT": "LS",  # Lesotho
    "LU": "LU",  # Luxembourg (same)
    "LY": "LY",  # Libya (same)
    "MA": "MG",  # Madagascar
    "MB": "MQ",  # Martinique
    "MC": "MO",  # Macau
    "MD": "MD",  # Moldova (same)
    "MF": "YT",  # Mayotte
    "MG": "MN",  # Mongolia
    "MH": "MS",  # Montserrat
    "MI": "MW",  # Malawi
    "MJ": "ME",  # Montenegro
    "MK": "MK",  # North Macedonia (same)
    "ML": "ML",  # Mali (same)
    "MM": "MM",  # Burma/Myanmar (same)
    "MN": "MC",  # Monaco
    "MO": "MA",  # Morocco
    "MP": "MU",  # Mauritius
    "MR": "MR",  # Mauritania (same)
    "MT": "MT",  # Malta (same)
    "MU": "OM",  # Oman
    "MV": "MV",  # Maldives (same)
    "MX": "MX",  # Mexico (same)
    "MY": "MY",  # Malaysia (same)
    "MZ": "MZ",  # Mozambique (same)
    "NC": "NC",  # New Caledonia (same)
    "NE": "NU",  # Niue
    "NF": "NF",  # Norfolk Island (same)
    "NG": "NE",  # Niger
    "NH": "VU",  # Vanuatu
    "NI": "NG",  # Nigeria
    "NL": "NL",  # Netherlands (same)
    "NO": "NO",  # Norway (same)
    "NP": "NP",  # Nepal (same)
    "NR": "NR",  # Nauru (same)
    "NS": "SR",  # Suriname
    "NU": "NI",  # Nicaragua
    "NZ": "NZ",  # New Zealand (same)
    "OD": "SS",  # South Sudan
    "PA": "PY",  # Paraguay
    "PC": "PN",  # Pitcairn Islands
    "PE": "PE",  # Peru (same)
    "PG": "PG",  # Spratly Islands / Papua New Guinea
    "PK": "PK",  # Pakistan (same)
    "PL": "PL",  # Poland (same)
    "PM": "PA",  # Panama
    "PO": "PT",  # Portugal
    "PP": "PG",  # Papua New Guinea
    "PS": "PW",  # Palau
    "PU": "GW",  # Guinea-Bissau
    "QA": "QA",  # Qatar (same)
    "RE": "RE",  # Reunion (same)
    "RI": "RS",  # Serbia
    "RM": "MH",  # Marshall Islands
    "RO": "RO",  # Romania (same)
    "RP": "PH",  # Philippines
    "RQ": "PR",  # Puerto Rico
    "RS": "RU",  # Russia
    "RW": "RW",  # Rwanda (same)
    "SA": "SA",  # Saudi Arabia (same)
    "SB": "PM",  # St. Pierre and Miquelon
    "SC": "KN",  # St. Kitts and Nevis
    "SE": "SC",  # Seychelles
    "SF": "ZA",  # South Africa
    "SG": "SN",  # Senegal
    "SH": "SH",  # St. Helena (same)
    "SI": "SI",  # Slovenia (same)
    "SL": "SL",  # Sierra Leone (same)
    "SM": "SM",  # San Marino (same)
    "SN": "SG",  # Singapore
    "SO": "SO",  # Somalia (same)
    "SP": "ES",  # Spain
    "SR": "RS",  # Serbia
    "ST": "LC",  # St. Lucia
    "SU": "SD",  # Sudan
    "SV": "SJ",  # Svalbard
    "SW": "SE",  # Sweden
    "SX": "GS",  # South Georgia
    "SY": "SY",  # Syria (same)
    "SZ": "CH",  # Switzerland
    "TC": "AE",  # United Arab Emirates
    "TD": "TT",  # Trinidad and Tobago
    "TH": "TH",  # Thailand (same)
    "TI": "TJ",  # Tajikistan
    "TK": "TC",  # Turks and Caicos Islands
    "TL": "TK",  # Tokelau
    "TN": "TO",  # Tonga
    "TO": "TG",  # Togo
    "TP": "ST",  # Sao Tome and Principe
    "TS": "TN",  # Tunisia
    "TU": "TR",  # Turkey
    "TV": "TV",  # Tuvalu (same)
    "TW": "TW",  # Taiwan (same)
    "TX": "TM",  # Turkmenistan
    "TZ": "TZ",  # Tanzania (same)
    "UG": "UG",  # Uganda (same)
    "UK": "GB",  # United Kingdom
    "UP": "UA",  # Ukraine
    "US": "US",  # United States (same)
    "UV": "BF",  # Burkina Faso
    "UY": "UY",  # Uruguay (same)
    "UZ": "UZ",  # Uzbekistan (same)
    "VC": "VC",  # St. Vincent and the Grenadines (same)
    "VE": "VE",  # Venezuela (same)
    "VI": "VG",  # Virgin Islands (British)
    "VM": "VN",  # Vietnam
    "VQ": "VI",  # Virgin Islands (U.S.)
    "VT": "VA",  # Vatican City
    "WA": "NA",  # Namibia
    "WE": "PS",  # West Bank (Palestine)
    "WF": "WF",  # Wallis and Futuna (same)
    "WI": "EH",  # Western Sahara
    "WQ": "UM",  # Wake Island (US Minor Outlying Islands)
    "WS": "WS",  # Western Samoa (same)
    "WZ": "SZ",  # Eswatini
    "YM": "YE",  # Yemen
    "ZA": "ZM",  # Zambia
    "ZI": "ZW",  # Zimbabwe
    "ZM": "WS",  # Samoa
}

# Reverse mapping: ISO to FIPS
ISO_TO_FIPS = {v: k for k, v in FIPS_TO_ISO.items()}

# ISO 3166-1 alpha-2 country codes to nicely-formatted country names
# Sourced from the gt package for standardized, display-friendly country names
ISO_COUNTRY_NAMES = {
    "AW": "Aruba",
    "AF": "Afghanistan",
    "AO": "Angola",
    "AL": "Albania",
    "AD": "Andorra",
    "AE": "United Arab Emirates",
    "AR": "Argentina",
    "AM": "Armenia",
    "AS": "American Samoa",
    "AG": "Antigua & Barbuda",
    "AU": "Australia",
    "AT": "Austria",
    "AZ": "Azerbaijan",
    "BI": "Burundi",
    "BE": "Belgium",
    "BJ": "Benin",
    "BF": "Burkina Faso",
    "BD": "Bangladesh",
    "BG": "Bulgaria",
    "BH": "Bahrain",
    "BS": "Bahamas",
    "BA": "Bosnia & Herzegovina",
    "BY": "Belarus",
    "BZ": "Belize",
    "BM": "Bermuda",
    "BO": "Bolivia",
    "BR": "Brazil",
    "BB": "Barbados",
    "BN": "Brunei",
    "BT": "Bhutan",
    "BW": "Botswana",
    "CF": "Central African Republic",
    "CA": "Canada",
    "CH": "Switzerland",
    "CL": "Chile",
    "CN": "China",
    "CI": "Côte d'Ivoire",
    "CM": "Cameroon",
    "CD": "Congo (DRC)",
    "CG": "Congo (Republic)",
    "CO": "Colombia",
    "KM": "Comoros",
    "CV": "Cape Verde",
    "CR": "Costa Rica",
    "CU": "Cuba",
    "CW": "Curaçao",
    "KY": "Cayman Islands",
    "CY": "Cyprus",
    "CZ": "Czech Republic",
    "DE": "Germany",
    "DJ": "Djibouti",
    "DM": "Dominica",
    "DK": "Denmark",
    "DO": "Dominican Republic",
    "DZ": "Algeria",
    "EC": "Ecuador",
    "EG": "Egypt",
    "ER": "Eritrea",
    "ES": "Spain",
    "EE": "Estonia",
    "ET": "Ethiopia",
    "FI": "Finland",
    "FJ": "Fiji",
    "FR": "France",
    "FO": "Faroe Islands",
    "FM": "Micronesia",
    "GA": "Gabon",
    "GB": "United Kingdom",
    "GE": "Georgia",
    "GH": "Ghana",
    "GI": "Gibraltar",
    "GN": "Guinea",
    "GM": "Gambia",
    "GW": "Guinea-Bissau",
    "GQ": "Equatorial Guinea",
    "GR": "Greece",
    "GD": "Grenada",
    "GL": "Greenland",
    "GT": "Guatemala",
    "GU": "Guam",
    "GY": "Guyana",
    "HK": "Hong Kong",
    "HN": "Honduras",
    "HR": "Croatia",
    "HT": "Haiti",
    "HU": "Hungary",
    "ID": "Indonesia",
    "IM": "Isle of Man",
    "IN": "India",
    "IE": "Ireland",
    "IR": "Iran",
    "IQ": "Iraq",
    "IS": "Iceland",
    "IL": "Israel",
    "IT": "Italy",
    "JM": "Jamaica",
    "JO": "Jordan",
    "JP": "Japan",
    "KZ": "Kazakhstan",
    "KE": "Kenya",
    "KG": "Kyrgyzstan",
    "KH": "Cambodia",
    "KI": "Kiribati",
    "KN": "St. Kitts & Nevis",
    "KR": "South Korea",
    "KW": "Kuwait",
    "LA": "Laos",
    "LB": "Lebanon",
    "LR": "Liberia",
    "LY": "Libya",
    "LC": "St. Lucia",
    "LI": "Liechtenstein",
    "LK": "Sri Lanka",
    "LS": "Lesotho",
    "LT": "Lithuania",
    "LU": "Luxembourg",
    "LV": "Latvia",
    "MO": "Macao",
    "MF": "St. Martin",
    "MA": "Morocco",
    "MC": "Monaco",
    "MD": "Moldova",
    "MG": "Madagascar",
    "MV": "Maldives",
    "MX": "Mexico",
    "MH": "Marshall Islands",
    "MK": "North Macedonia",
    "ML": "Mali",
    "MT": "Malta",
    "MM": "Myanmar",
    "ME": "Montenegro",
    "MN": "Mongolia",
    "MP": "Northern Mariana Islands",
    "MZ": "Mozambique",
    "MR": "Mauritania",
    "MU": "Mauritius",
    "MW": "Malawi",
    "MY": "Malaysia",
    "NA": "Namibia",
    "NC": "New Caledonia",
    "NE": "Niger",
    "NG": "Nigeria",
    "NI": "Nicaragua",
    "NL": "Netherlands",
    "NO": "Norway",
    "NP": "Nepal",
    "NR": "Nauru",
    "NZ": "New Zealand",
    "OM": "Oman",
    "PK": "Pakistan",
    "PA": "Panama",
    "PE": "Peru",
    "PH": "Philippines",
    "PW": "Palau",
    "PG": "Papua New Guinea",
    "PL": "Poland",
    "PR": "Puerto Rico",
    "KP": "North Korea",
    "PT": "Portugal",
    "PY": "Paraguay",
    "PS": "Palestine",
    "PF": "French Polynesia",
    "QA": "Qatar",
    "RO": "Romania",
    "RU": "Russia",
    "RW": "Rwanda",
    "SA": "Saudi Arabia",
    "SD": "Sudan",
    "SN": "Senegal",
    "SG": "Singapore",
    "SB": "Solomon Islands",
    "SL": "Sierra Leone",
    "SV": "El Salvador",
    "SM": "San Marino",
    "SO": "Somalia",
    "RS": "Serbia",
    "SS": "South Sudan",
    "ST": "São Tomé & Príncipe",
    "SR": "Suriname",
    "SK": "Slovakia",
    "SI": "Slovenia",
    "SE": "Sweden",
    "SZ": "Eswatini",
    "SX": "Sint Maarten",
    "SC": "Seychelles",
    "SY": "Syria",
    "TC": "Turks & Caicos Islands",
    "TD": "Chad",
    "TG": "Togo",
    "TH": "Thailand",
    "TJ": "Tajikistan",
    "TM": "Turkmenistan",
    "TL": "East Timor",
    "TO": "Tonga",
    "TT": "Trinidad & Tobago",
    "TN": "Tunisia",
    "TR": "Turkey",
    "TV": "Tuvalu",
    "TZ": "Tanzania",
    "UG": "Uganda",
    "UA": "Ukraine",
    "UY": "Uruguay",
    "US": "United States",
    "UZ": "Uzbekistan",
    "VC": "St. Vincent & Grenadines",
    "VE": "Venezuela",
    "VG": "British Virgin Islands",
    "VI": "U.S. Virgin Islands",
    "VN": "Vietnam",
    "VU": "Vanuatu",
    "WS": "Samoa",
    "YE": "Yemen",
    "ZA": "South Africa",
    "ZM": "Zambia",
    "ZW": "Zimbabwe",
}


# US state and territory codes to names
US_STATE_NAMES = {
    "AL": "Alabama",
    "AK": "Alaska",
    "AZ": "Arizona",
    "AR": "Arkansas",
    "CA": "California",
    "CO": "Colorado",
    "CT": "Connecticut",
    "DE": "Delaware",
    "FL": "Florida",
    "GA": "Georgia",
    "HI": "Hawaii",
    "ID": "Idaho",
    "IL": "Illinois",
    "IN": "Indiana",
    "IA": "Iowa",
    "KS": "Kansas",
    "KY": "Kentucky",
    "LA": "Louisiana",
    "ME": "Maine",
    "MD": "Maryland",
    "MA": "Massachusetts",
    "MI": "Michigan",
    "MN": "Minnesota",
    "MS": "Mississippi",
    "MO": "Missouri",
    "MT": "Montana",
    "NE": "Nebraska",
    "NV": "Nevada",
    "NH": "New Hampshire",
    "NJ": "New Jersey",
    "NM": "New Mexico",
    "NY": "New York",
    "NC": "North Carolina",
    "ND": "North Dakota",
    "OH": "Ohio",
    "OK": "Oklahoma",
    "OR": "Oregon",
    "PA": "Pennsylvania",
    "RI": "Rhode Island",
    "SC": "South Carolina",
    "SD": "South Dakota",
    "TN": "Tennessee",
    "TX": "Texas",
    "UT": "Utah",
    "VT": "Vermont",
    "VA": "Virginia",
    "WA": "Washington",
    "WV": "West Virginia",
    "WI": "Wisconsin",
    "WY": "Wyoming",
    "DC": "District of Columbia",
    "AS": "American Samoa",
    "GU": "Guam",
    "MP": "Northern Mariana Islands",
    "PR": "Puerto Rico",
    "VI": "U.S. Virgin Islands",
}


class _CountryCodes:
    """
    Convenience class for accessing ISO country codes via attribute access.

    Provides a clean syntax for specifying country codes in function calls:
        wv.search_stations(country_code=wv.country.DE)
        wv.search_stations(country_code=wv.country.GB)

    All ISO 3166-1 alpha-2 country codes are available as attributes.
    """

    def __getattr__(self, name: str) -> str:
        """Return the ISO country code (same as the attribute name)."""
        code = name.upper()
        if code in ISO_COUNTRY_NAMES:
            return code
        raise AttributeError(
            f"'{code}' is not a valid ISO 3166-1 alpha-2 country code. "
            f"See wv.ISO_COUNTRY_NAMES for valid codes."
        )

    def __dir__(self):
        """Return list of all valid ISO country codes for autocomplete."""
        return list(ISO_COUNTRY_NAMES.keys())

    def __repr__(self) -> str:
        return f"<CountryCodes: {len(ISO_COUNTRY_NAMES)} ISO 3166-1 alpha-2 codes>"


# Singleton instance for convenient attribute-based country code access
country = _CountryCodes()


class _StateCodes:
    """
    Convenience class for accessing US state and territory codes via attribute access.

    Provides a clean syntax for specifying state codes in function calls:
        wv.search_stations(state=wv.state.CA)
        wv.search_stations(state=wv.state.NY)
        wv.search_stations(state=wv.state.TX)

    All US state and territory codes are available as attributes.
    Includes the 50 states, DC, and 5 territories (AS, GU, MP, PR, VI).
    """

    def __getattr__(self, name: str) -> str:
        """Return the US state code (same as the attribute name)."""
        code = name.upper()
        if code in US_STATE_NAMES:
            return code
        raise AttributeError(
            f"'{code}' is not a valid US state or territory code. "
            f"See wv.US_STATE_NAMES for valid codes."
        )

    def __dir__(self):
        """Return list of all valid US state codes for autocomplete."""
        return list(US_STATE_NAMES.keys())

    def __repr__(self) -> str:
        return f"<StateCodes: {len(US_STATE_NAMES)} US states and territories>"


# Singleton instance for convenient attribute-based state code access
state = _StateCodes()



"""ISD data format parsing utilities."""

from __future__ import annotations

import gzip
import io
from typing import TYPE_CHECKING

import polars as pl

from weathervault._constants import (
    MANDATORY_COLUMN_NAMES,
    MANDATORY_COLUMN_WIDTHS,
    MISSING_VALUES,
    SCALE_FACTORS,
)

if TYPE_CHECKING:
    pass


def parse_isd_line(line: str) -> dict:
    """
    Parse a single line of ISD data into a dictionary.

    Parameters
    ----------
    line
        A single line from an ISD data file.

    Returns
    -------
    dict
        Dictionary containing parsed values from the mandatory data section.
    """
    values = {}
    pos = 0

    for _i, (name, width) in enumerate(
        zip(MANDATORY_COLUMN_NAMES, MANDATORY_COLUMN_WIDTHS, strict=True)
    ):
        if pos + width > len(line):
            # Line is shorter than expected, fill with None
            values[name] = None
        else:
            raw_value = line[pos : pos + width].strip()
            values[name] = raw_value if raw_value else None
        pos += width

    return values


def parse_isd_data(data: bytes | str) -> pl.DataFrame:
    """
    Parse ISD format data (gzipped or plain text) into a Polars DataFrame.

    Parameters
    ----------
    data
        The raw ISD data as bytes (possibly gzipped) or string.

    Returns
    -------
    pl.DataFrame
        DataFrame containing the parsed weather observations.
    """
    # Handle gzipped data
    if isinstance(data, bytes):
        try:
            # Try to decompress as gzip
            with gzip.GzipFile(fileobj=io.BytesIO(data)) as f:
                text = f.read().decode("utf-8", errors="replace")
        except gzip.BadGzipFile:
            # Not gzipped, decode directly
            text = data.decode("utf-8", errors="replace")
    else:
        text = data

    lines = text.strip().split("\n")
    if not lines:  # pragma: no cover
        return _empty_weather_dataframe()

    # Parse each line
    records = []
    for line in lines:
        if line.strip():
            record = parse_isd_line(line)
            records.append(record)

    if not records:
        return _empty_weather_dataframe()

    # Create DataFrame from records
    df = pl.DataFrame(records)

    return df


def process_weather_data(
    df: pl.DataFrame,
    *,
    tz_name: str | None = None,
    convert_to_local: bool = True,
    temp_unit: str = "celsius",
    stn_metadata: dict | None = None,
) -> pl.DataFrame:
    """
    Process raw parsed ISD data into a clean weather DataFrame.

    Parameters
    ----------
    df
        Raw DataFrame from parse_isd_data.
    tz_name
        Timezone name for the station (e.g., "America/New_York").
        Used for converting UTC times to local time.
    convert_to_local
        Whether to convert times from UTC to local time. Default is True.
    temp_unit
        Temperature unit for output: "celsius" or "fahrenheit". Default is "celsius".
    stn_metadata
        Optional dictionary containing station metadata to include in output.
        Should contain keys: name, country, state, icao, lat, lon, elev.

    Returns
    -------
    pl.DataFrame
        Processed DataFrame with clean, usable weather data.
    """
    if df.height == 0:
        return _empty_processed_dataframe()

    # Convert numeric columns
    df = df.with_columns(
        [
            pl.col("year").cast(pl.Int32, strict=False),
            pl.col("month").cast(pl.Int32, strict=False),
            pl.col("day").cast(pl.Int32, strict=False),
            pl.col("hour").cast(pl.Int32, strict=False),
            pl.col("minute").cast(pl.Int32, strict=False),
            pl.col("wind_direction").cast(pl.Int32, strict=False),
            pl.col("wind_speed").cast(pl.Int32, strict=False),
            pl.col("ceiling_height").cast(pl.Int32, strict=False),
            pl.col("visibility").cast(pl.Int32, strict=False),
            pl.col("temp").cast(pl.Int32, strict=False),
            pl.col("dew_point").cast(pl.Int32, strict=False),
            pl.col("sea_level_pressure").cast(pl.Int32, strict=False),
        ]
    )

    # Create station ID
    df = df.with_columns(pl.concat_str([pl.col("usaf"), pl.lit("-"), pl.col("wban")]).alias("id"))

    # Create datetime column (UTC)
    df = df.with_columns(
        pl.datetime(
            year=pl.col("year"),
            month=pl.col("month"),
            day=pl.col("day"),
            hour=pl.col("hour"),
            minute=pl.col("minute"),
            time_zone="UTC",
        ).alias("time_utc")
    )

    # Replace missing values with null
    df = df.with_columns(
        [
            pl.when(pl.col("wind_direction") == MISSING_VALUES["wind_direction"])
            .then(None)
            .otherwise(pl.col("wind_direction"))
            .alias("wind_direction"),
            pl.when(pl.col("wind_speed") == MISSING_VALUES["wind_speed"])
            .then(None)
            .otherwise(pl.col("wind_speed"))
            .alias("wind_speed"),
            pl.when(pl.col("ceiling_height") == MISSING_VALUES["ceiling_height"])
            .then(None)
            .otherwise(pl.col("ceiling_height"))
            .alias("ceiling_height"),
            pl.when(pl.col("visibility") == MISSING_VALUES["visibility"])
            .then(None)
            .otherwise(pl.col("visibility"))
            .alias("visibility"),
            pl.when(pl.col("temp") == MISSING_VALUES["temp"])
            .then(None)
            .otherwise(pl.col("temp"))
            .alias("temp"),
            pl.when(pl.col("dew_point") == MISSING_VALUES["dew_point"])
            .then(None)
            .otherwise(pl.col("dew_point"))
            .alias("dew_point"),
            pl.when(pl.col("sea_level_pressure") == MISSING_VALUES["sea_level_pressure"])
            .then(None)
            .otherwise(pl.col("sea_level_pressure"))
            .alias("sea_level_pressure"),
        ]
    )

    # Apply scale factors
    df = df.with_columns(
        [
            (pl.col("wind_speed").cast(pl.Float64) / SCALE_FACTORS["wind_speed"]).alias(
                "wind_speed"
            ),
            (pl.col("temp").cast(pl.Float64) / SCALE_FACTORS["temp"]).alias("temp"),
            (pl.col("dew_point").cast(pl.Float64) / SCALE_FACTORS["dew_point"]).alias("dew_point"),
            (
                pl.col("sea_level_pressure").cast(pl.Float64) / SCALE_FACTORS["sea_level_pressure"]
            ).alias("sea_level_pressure"),
        ]
    )

    # Calculate relative humidity using the August-Roche-Magnus approximation
    df = df.with_columns(
        (
            100.0
            * (
                ((17.625 * pl.col("dew_point")) / (243.04 + pl.col("dew_point"))).exp()
                / ((17.625 * pl.col("temp")) / (243.04 + pl.col("temp"))).exp()
            )
        )
        .round(1)
        .alias("relative_humidity")
    )

    # Convert temperature if requested (data is in Celsius by default)
    if temp_unit == "fahrenheit":
        df = df.with_columns(
            [
                ((pl.col("temp") * 9 / 5) + 32).round(1).alias("temp"),
                ((pl.col("dew_point") * 9 / 5) + 32).round(1).alias("dew_point"),
            ]
        )
    elif temp_unit == "kelvin":
        df = df.with_columns(
            [
                (pl.col("temp") + 273.15).round(2).alias("temp"),
                (pl.col("dew_point") + 273.15).round(2).alias("dew_point"),
            ]
        )

    # Handle time conversion to local
    if convert_to_local and tz_name:
        # Convert UTC datetime to local time
        df = df.with_columns(pl.col("time_utc").dt.convert_time_zone(tz_name).alias("time"))
    else:
        df = df.with_columns(pl.col("time_utc").alias("time"))

    # Build column list for final selection
    columns = [
        pl.col("id"),
        pl.col("time"),
        pl.col("temp"),
        pl.col("dew_point"),
        pl.col("relative_humidity").alias("rh"),
        pl.col("wind_direction").alias("wd"),
        pl.col("wind_speed").alias("ws"),
        pl.col("sea_level_pressure").alias("atmos_pres"),
        pl.col("ceiling_height").alias("ceil_hgt"),
        pl.col("visibility"),
    ]

    # Add station metadata columns if provided
    if stn_metadata is not None:
        df = df.with_columns(
            [
                pl.lit(stn_metadata["name"]).alias("name"),
                pl.lit(stn_metadata["country"]).alias("country"),
                pl.lit(stn_metadata["state"]).alias("state"),
                pl.lit(stn_metadata["icao"]).alias("icao"),
                pl.lit(stn_metadata["lat"]).alias("lat"),
                pl.lit(stn_metadata["lon"]).alias("lon"),
                pl.lit(stn_metadata["elev"]).alias("elev"),
            ]
        )
        columns.extend(
            [
                pl.col("name"),
                pl.col("country"),
                pl.col("state"),
                pl.col("icao"),
                pl.col("lat"),
                pl.col("lon"),
                pl.col("elev"),
            ]
        )

    # Select and rename final columns
    result = df.select(columns)

    return result


def _empty_weather_dataframe() -> pl.DataFrame:
    """Create an empty DataFrame with the raw ISD column schema."""
    return pl.DataFrame(schema=dict.fromkeys(MANDATORY_COLUMN_NAMES, pl.Utf8))


def _empty_processed_dataframe(
    incl_stn_info: bool = False, time_as_columns: bool = False
) -> pl.DataFrame:
    """Create an empty DataFrame with the processed weather schema.

    Parameters
    ----------
    incl_stn_info
        Whether to include station metadata columns in the schema.
    time_as_columns
        Whether to use separate time component columns instead of a single time column.
    """
    if time_as_columns:
        schema = {
            "id": pl.Utf8,
            "year": pl.Int32,
            "month": pl.Int32,
            "day": pl.Int32,
            "hour": pl.Int32,
            "min": pl.Int32,
            "temp": pl.Float64,
            "dew_point": pl.Float64,
            "rh": pl.Float64,
            "wd": pl.Int32,
            "ws": pl.Float64,
            "atmos_pres": pl.Float64,
            "ceil_hgt": pl.Int32,
            "visibility": pl.Int32,
        }
    else:
        schema = {
            "id": pl.Utf8,
            "time": pl.Datetime("us", "UTC"),
            "temp": pl.Float64,
            "dew_point": pl.Float64,
            "rh": pl.Float64,
            "wd": pl.Int32,
            "ws": pl.Float64,
            "atmos_pres": pl.Float64,
            "ceil_hgt": pl.Int32,
            "visibility": pl.Int32,
        }

    if incl_stn_info:
        schema.update(
            {
                "name": pl.Utf8,
                "country": pl.Utf8,
                "state": pl.Utf8,
                "icao": pl.Utf8,
                "lat": pl.Float64,
                "lon": pl.Float64,
                "elev": pl.Float64,
            }
        )

    return pl.DataFrame(schema=schema)



"""Tests for the stations module."""

import polars as pl
import pytest

import weathervault as wv
from weathervault._constants import COUNTRY_CODES
from weathervault.stations import (
    get_countries,
    get_inventory,
    get_station_metadata,
    get_years_for_station,
    search_stations,
)


class TestGetCountries:
    """Tests for get_countries function."""

    def test_returns_dataframe(self):
        """Test that get_countries returns a DataFrame."""
        result = get_countries()

        assert isinstance(result, pl.DataFrame)
        assert result.height > 0

    def test_has_expected_columns(self):
        """Test that result has expected columns."""
        result = get_countries()

        assert "country_code" in result.columns
        assert "country" in result.columns

    def test_sorted_by_country(self):
        """Test that results are sorted by country name."""
        result = get_countries()

        countries = result["country"].to_list()
        assert countries == sorted(countries)

    def test_contains_known_countries(self):
        """Test that known countries are present."""
        result = get_countries()
        countries = result["country"].to_list()

        assert "United States" in countries
        assert "Canada" in countries
        assert "United Kingdom" in countries


class TestGetStationMetadata:
    """Tests for get_station_metadata function.

    Note: These tests require network access to download station data.
    """

    @pytest.mark.network
    def test_returns_dataframe(self):
        """Test that get_station_metadata returns a DataFrame."""
        result = get_station_metadata()

        assert isinstance(result, pl.DataFrame)
        assert result.height > 0

    @pytest.mark.network
    def test_has_expected_columns(self):
        """Test that result has expected columns."""
        result = get_station_metadata()

        expected_columns = [
            "id",
            "usaf",
            "wban",
            "name",
            "country_code",
            "country",
            "state",
            "icao",
            "lat",
            "lon",
            "elev",
            "begin_date",
            "end_date",
            "tz_name",
        ]
        for col in expected_columns:
            assert col in result.columns

    @pytest.mark.network
    def test_id_format(self):
        """Test that station IDs are in USAF-WBAN format."""
        result = get_station_metadata()

        # Check first few IDs contain a hyphen
        for station_id in result["id"].head(10).to_list():
            assert "-" in station_id

    @pytest.mark.network
    def test_caching(self):
        """Test that results are cached."""
        result1 = get_station_metadata()
        result2 = get_station_metadata()

        # Should return the same DataFrame object (cached)
        assert result1 is result2

    @pytest.mark.network
    def test_force_refresh(self):
        """Test that force_refresh bypasses cache."""
        result1 = get_station_metadata()
        result2 = get_station_metadata(force_refresh=True)

        # Should return different DataFrame objects
        # (content should be same but different instance)
        assert result1 is not result2
        assert result1.height == result2.height

    @pytest.mark.network
    def test_cache_with_timezone_then_without(self):
        """Test cache behavior: fetch with timezone, then without."""
        # First fetch with timezone (populates cache with tz)
        result_with_tz = get_station_metadata(include_timezone=True, force_refresh=True)
        assert "tz_name" in result_with_tz.columns

        # Now fetch without timezone - should use cache and drop tz_name
        result_without_tz = get_station_metadata(include_timezone=False)
        assert "tz_name" not in result_without_tz.columns
        assert result_with_tz.height == result_without_tz.height

    @pytest.mark.network
    def test_cache_without_timezone_reused(self):
        """Test cache reuse when fetched without timezone."""
        # Clear cache by forcing refresh without timezone
        import weathervault.stations as stations_module

        stations_module._station_metadata_cache = None
        stations_module._station_metadata_cache_has_tz = False

        # Fetch without timezone
        result1 = get_station_metadata(include_timezone=False)
        assert "tz_name" not in result1.columns

        # Fetch again without timezone - should use cache
        result2 = get_station_metadata(include_timezone=False)
        assert result1 is result2


class TestSearchStations:
    """Tests for search_stations function."""

    @pytest.mark.network
    def test_search_by_state(self):
        """Test searching stations by US state."""
        result = search_stations(state="CA")

        assert isinstance(result, pl.DataFrame)
        # All results should be in California
        assert all(state == "CA" for state in result["state"].to_list() if state)

    @pytest.mark.network
    def test_search_by_country(self):
        """Test searching stations by country name."""
        result = search_stations(country="Canada")

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            assert all(code == "CA" for code in result["country_code"].to_list() if code)

    @pytest.mark.network
    def test_search_by_country_code(self):
        """Test searching stations by country code."""
        result = search_stations(country_code="UK")

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            assert all(code == "UK" for code in result["country_code"].to_list())

    @pytest.mark.network
    def test_search_by_lat_range(self):
        """Test searching stations by latitude range."""
        result = search_stations(lat_range=(40.0, 41.0))

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            lats = result["lat"].to_list()
            assert all(40.0 <= lat <= 41.0 for lat in lats if lat is not None)

    @pytest.mark.network
    def test_search_by_lon_range(self):
        """Test searching stations by longitude range."""
        result = search_stations(lon_range=(-75.0, -74.0))

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            lons = result["lon"].to_list()
            assert all(-75.0 <= lon <= -74.0 for lon in lons if lon is not None)

    @pytest.mark.network
    def test_search_by_name(self):
        """Test searching stations by name."""
        result = search_stations(name="airport")

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            names = result["name"].to_list()
            assert all("airport" in name.lower() for name in names if name)

    @pytest.mark.network
    def test_combined_search(self):
        """Test searching with multiple criteria."""
        result = search_stations(state="NY", name="airport")

        assert isinstance(result, pl.DataFrame)

    @pytest.mark.network
    def test_search_has_recent_data(self):
        """Test filtering stations by recent data availability."""
        from datetime import date

        result = search_stations(has_recent_data=True)

        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            # All stations should have end_date >= last year
            current_year = date.today().year
            cutoff = date(current_year - 1, 1, 1)
            end_dates = result["end_date"].to_list()
            assert all(d >= cutoff for d in end_dates if d is not None)


class TestGetInventory:
    """Tests for get_inventory function."""

    @pytest.mark.network
    def test_returns_dataframe(self):
        """Test that get_inventory returns a DataFrame."""
        result = get_inventory()

        assert isinstance(result, pl.DataFrame)
        assert result.height > 0

    @pytest.mark.network
    def test_has_expected_columns(self):
        """Test that result has expected columns."""
        result = get_inventory()

        expected_columns = [
            "id",
            "usaf",
            "wban",
            "year",
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
            "total",
        ]
        for col in expected_columns:
            assert col in result.columns

    @pytest.mark.network
    def test_total_calculated(self):
        """Test that total column is sum of monthly values."""
        result = get_inventory()

        # Check first row
        row = result.row(0, named=True)
        expected_total = sum(
            [
                row["jan"],
                row["feb"],
                row["mar"],
                row["apr"],
                row["may"],
                row["jun"],
                row["jul"],
                row["aug"],
                row["sep"],
                row["oct"],
                row["nov"],
                row["dec"],
            ]
        )
        assert row["total"] == expected_total


class TestGetYearsForStation:
    """Tests for get_years_for_station function."""

    @pytest.mark.network
    def test_returns_list(self):
        """Test that function returns a list."""
        # Use a known station (LaGuardia Airport)
        result = get_years_for_station("725030-14732")

        assert isinstance(result, list)

    @pytest.mark.network
    def test_years_sorted(self):
        """Test that years are sorted."""
        result = get_years_for_station("725030-14732")

        if result:
            assert result == sorted(result)

    @pytest.mark.network
    def test_invalid_station(self):
        """Test that invalid station returns empty list."""
        result = get_years_for_station("000000-00000")

        assert result == []

    @pytest.mark.network
    def test_known_station_has_recent_years(self):
        """Test that a known active station has recent years."""
        result = get_years_for_station("725030-14732")
        # LaGuardia should have data at least from 2020 onwards
        assert any(year >= 2020 for year in result)

    @pytest.mark.network
    def test_known_station_has_historical_years(self):
        """Test that a known station has historical data."""
        result = get_years_for_station("725030-14732")
        # LaGuardia has been recording for decades
        assert any(year < 2000 for year in result)


class TestCountryCodeMapping:
    """Tests for country code mapping in search and metadata."""

    def test_country_codes_dict_not_empty(self):
        """Test that COUNTRY_CODES dictionary is not empty."""
        assert len(COUNTRY_CODES) > 0

    def test_common_fips_codes(self):
        """Test some common FIPS country codes."""
        assert COUNTRY_CODES["US"] == "United States"
        assert COUNTRY_CODES["CA"] == "Canada"
        assert COUNTRY_CODES["UK"] == "United Kingdom"
        assert COUNTRY_CODES["AS"] == "Australia"  # FIPS code for Australia
        assert COUNTRY_CODES["FR"] == "France"

    @pytest.mark.network
    def test_search_by_country_uses_mapping(self):
        """Test that search uses country name to code mapping."""
        # Search by full country name should work
        result = search_stations(country="Australia")
        if result.height > 0:
            # Should find stations with ISO code 'AU' (not FIPS 'AS')
            assert all(code == "AU" for code in result["country_code"].to_list() if code)


class TestSearchStationsAdvanced:
    """Advanced tests for search_stations function."""

    @pytest.mark.network
    def test_empty_result_for_impossible_criteria(self):
        """Test that impossible criteria returns empty DataFrame."""
        # Non-existent country code
        result = search_stations(country_code="ZZ")
        assert result.height == 0

    @pytest.mark.network
    def test_latitude_longitude_combined(self):
        """Test combined latitude and longitude search."""
        # Around New York City
        result = search_stations(lat_range=(40.5, 41.0), lon_range=(-74.5, -73.5))
        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            # Verify all results are in bounds
            for row in result.iter_rows(named=True):
                if row["lat"] is not None:
                    assert 40.5 <= row["lat"] <= 41.0
                if row["lon"] is not None:
                    assert -74.5 <= row["lon"] <= -73.5

    @pytest.mark.network
    def test_name_search_case_insensitive(self):
        """Test that name search is case insensitive."""
        upper_result = search_stations(name="AIRPORT")
        lower_result = search_stations(name="airport")
        # Both should return results
        assert upper_result.height > 0
        assert lower_result.height > 0

    @pytest.mark.network
    def test_state_combined_with_lat_range(self):
        """Test combining state with latitude range."""
        result = search_stations(state="CA", lat_range=(35.0, 40.0))
        assert isinstance(result, pl.DataFrame)
        if result.height > 0:
            # All should be in California and within lat range
            for row in result.iter_rows(named=True):
                if row["state"]:
                    assert row["state"] == "CA"
                if row["lat"] is not None:
                    assert 35.0 <= row["lat"] <= 40.0


class TestStationMetadataDetails:
    """Detailed tests for station metadata quality."""

    @pytest.mark.network
    def test_timezone_populated(self):
        """Test that timezone is populated for stations with coordinates."""
        result = get_station_metadata()
        # Stations with lat/lon should have timezone
        with_coords = result.filter(pl.col("lat").is_not_null() & pl.col("lon").is_not_null())
        if with_coords.height > 0:
            # At least some should have timezone
            tz_populated = with_coords.filter(pl.col("tz_name").is_not_null()).height
            assert tz_populated > 0

    @pytest.mark.network
    def test_station_dates_valid(self):
        """Test that begin_date and end_date are valid dates."""
        result = get_station_metadata()
        # Check that dates are not null for at least some stations
        with_dates = result.filter(pl.col("begin_date").is_not_null())
        assert with_dates.height > 0

    @pytest.mark.network
    def test_elevation_reasonable(self):
        """Test that elevations are in reasonable range."""
        result = get_station_metadata()
        elevations = result.filter(pl.col("elev").is_not_null())["elev"].to_list()
        # Elevations should be between -500m (Dead Sea area) and 10000m (extreme mountains)
        # Missing values (-999.0, -999.9) should already be converted to null
        out_of_range = [elev for elev in elevations if not (-500 < elev < 10000)]
        assert len(out_of_range) == 0, (
            f"Found {len(out_of_range)} elevations out of range [-500, 10000]: "
            f"min={min(out_of_range) if out_of_range else 'N/A'}, "
            f"max={max(out_of_range) if out_of_range else 'N/A'}"
        )


class TestInventoryDetails:
    """Detailed tests for inventory data."""

    @pytest.mark.network
    def test_monthly_values_non_negative(self):
        """Test that monthly observation counts are non-negative."""
        result = get_inventory()
        months = [
            "jan",
            "feb",
            "mar",
            "apr",
            "may",
            "jun",
            "jul",
            "aug",
            "sep",
            "oct",
            "nov",
            "dec",
        ]
        for month in months:
            assert all(v >= 0 for v in result[month].to_list())

    @pytest.mark.network
    def test_years_reasonable(self):
        """Test that years in inventory are reasonable."""
        result = get_inventory()
        years = result["year"].unique().to_list()
        # Should have years from historical to recent
        assert min(years) < 2000  # Has historical data
        assert max(years) >= 2020  # Has recent data

    @pytest.mark.network
    def test_inventory_matches_station_format(self):
        """Test that inventory IDs match station ID format."""
        result = get_inventory()
        # All IDs should contain hyphen
        assert all("-" in sid for sid in result["id"].to_list())


class TestCountryShorthand:
    """Tests for the country code shorthand (wv.country.XX)."""

    def test_country_exists(self):
        """Test that wv.country exists."""
        assert hasattr(wv, "country")

    def test_returns_uppercase_code(self):
        """Test that country codes are returned in uppercase."""
        assert wv.country.US == "US"
        assert wv.country.GB == "GB"
        assert wv.country.DE == "DE"
        assert wv.country.FR == "FR"

    def test_lowercase_access_works(self):
        """Test that lowercase attribute access works."""
        assert wv.country.us == "US"
        assert wv.country.gb == "GB"
        assert wv.country.de == "DE"

    def test_invalid_code_raises_error(self):
        """Test that invalid country codes raise AttributeError."""
        with pytest.raises(AttributeError, match="not a valid ISO 3166-1 alpha-2"):
            _ = wv.country.INVALID

    @pytest.mark.network
    def test_works_in_search_stations(self):
        """Test that country shorthand works in search_stations."""
        # Should work just like passing string "US"
        result1 = search_stations(country_code=wv.country.US)
        result2 = search_stations(country_code="US")

        assert result1.height == result2.height
        assert result1.height > 0

    @pytest.mark.network
    def test_multiple_countries(self):
        """Test using country shorthand for multiple countries."""
        de = search_stations(country_code=wv.country.DE)
        gb = search_stations(country_code=wv.country.GB)
        fr = search_stations(country_code=wv.country.FR)

        assert de.height > 0
        assert gb.height > 0
        assert fr.height > 0

        # Verify they return different results
        assert de.height != gb.height
        assert de.height != fr.height

    def test_dir_returns_country_codes(self):
        """Test that dir() returns list of country codes."""
        codes = dir(wv.country)

        # Should contain known ISO codes
        assert "US" in codes
        assert "GB" in codes
        assert "DE" in codes
        assert "FR" in codes
        assert len(codes) > 200  # Should have many country codes

    def test_repr(self):
        """Test string representation of country object."""
        repr_str = repr(wv.country)
        assert "CountryCodes" in repr_str
        assert "ISO 3166-1" in repr_str


class TestStateShorthand:
    """Tests for the US state code shorthand (wv.state.XX)."""

    def test_state_exists(self):
        """Test that wv.state exists."""
        assert hasattr(wv, "state")

    def test_returns_uppercase_code(self):
        """Test that state codes are returned in uppercase."""
        assert wv.state.CA == "CA"
        assert wv.state.NY == "NY"
        assert wv.state.TX == "TX"
        assert wv.state.FL == "FL"

    def test_lowercase_access_works(self):
        """Test that lowercase attribute access works."""
        assert wv.state.ca == "CA"
        assert wv.state.ny == "NY"
        assert wv.state.tx == "TX"

    def test_territories_included(self):
        """Test that US territories are included."""
        assert wv.state.PR == "PR"  # Puerto Rico
        assert wv.state.GU == "GU"  # Guam
        assert wv.state.VI == "VI"  # U.S. Virgin Islands
        assert wv.state.AS == "AS"  # American Samoa
        assert wv.state.MP == "MP"  # Northern Mariana Islands
        assert wv.state.DC == "DC"  # District of Columbia

    def test_invalid_code_raises_error(self):
        """Test that invalid state codes raise AttributeError."""
        with pytest.raises(AttributeError, match="not a valid US state"):
            _ = wv.state.INVALID

    @pytest.mark.network
    def test_works_in_search_stations(self):
        """Test that state shorthand works in search_stations."""
        # Should work just like passing string "CA"
        result1 = search_stations(state=wv.state.CA)
        result2 = search_stations(state="CA")

        assert result1.height == result2.height
        assert result1.height > 0

    @pytest.mark.network
    def test_multiple_states(self):
        """Test using state shorthand for multiple states."""
        ca = search_stations(state=wv.state.CA)
        ny = search_stations(state=wv.state.NY)
        tx = search_stations(state=wv.state.TX)

        assert ca.height > 0
        assert ny.height > 0
        assert tx.height > 0

        # Verify they return different results
        assert ca.height != ny.height
        assert ca.height != tx.height

    def test_dir_returns_state_codes(self):
        """Test that dir() returns list of state codes."""
        codes = dir(wv.state)

        # Should contain known state codes
        assert "CA" in codes
        assert "NY" in codes
        assert "TX" in codes
        assert "FL" in codes
        assert len(codes) == 56  # 50 states + DC + 5 territories

    def test_repr(self):
        """Test string representation of state object."""
        repr_str = repr(wv.state)
        assert "StateCodes" in repr_str
        assert "US states" in repr_str



"""Station registry for IDE autocomplete support."""

from __future__ import annotations

import re
from functools import lru_cache
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    import polars as pl


def _sanitize_name(name: str) -> str:
    """Convert station name to valid Python identifier.

    Examples:
        "NEW YORK LAGUARDIA AP" -> "NEW_YORK_LAGUARDIA_AP"
        "JOHN F KENNEDY INTL" -> "JOHN_F_KENNEDY_INTL"
        "ST. LOUIS" -> "ST_LOUIS"
    """
    if not name:
        return "UNKNOWN"
    # Replace non-alphanumeric with underscore, collapse multiple underscores
    sanitized = re.sub(r"[^A-Z0-9]+", "_", name.upper())
    # Remove leading/trailing underscores
    sanitized = sanitized.strip("_")
    # Ensure it doesn't start with a number
    if sanitized and sanitized[0].isdigit():
        sanitized = "_" + sanitized
    return sanitized or "UNKNOWN"


@lru_cache(maxsize=1)
def _get_stations_df() -> pl.DataFrame:
    """Get station metadata (cached)."""
    from weathervault.stations import get_station_metadata

    return get_station_metadata(include_timezone=False)


class _StationNamespace:
    """A namespace that provides station IDs with autocomplete support.

    This class lazily loads station data and provides attribute access to station IDs organized by
    country and optionally state.
    """

    def __init__(
        self,
        stations: dict[str, str] | None = None,
        children: dict[str, _StationNamespace] | None = None,
    ):
        """Initialize namespace.

        Args:
            stations: Dict mapping sanitized names to station IDs
            children: Dict mapping child namespace names to child namespaces
        """
        self._stations = stations or {}
        self._children = children or {}

    def __getattr__(self, name: str) -> _StationNamespace | str:
        """Get a child namespace or station ID."""
        if name.startswith("_"):
            raise AttributeError(name)

        # Check children first (country codes, states)
        if name in self._children:
            return self._children[name]

        # Then check stations
        if name in self._stations:
            return self._stations[name]

        # Try case-insensitive match
        name_upper = name.upper()
        for key, value in self._children.items():
            if key.upper() == name_upper:
                return value
        for key, value in self._stations.items():
            if key.upper() == name_upper:
                return value

        raise AttributeError(
            f"No station or group named '{name}'. Use dir() to see available options."
        )

    def __dir__(self) -> list[str]:
        """Return list of available attributes for autocomplete."""
        return sorted(set(list(self._children.keys()) + list(self._stations.keys())))

    def __repr__(self) -> str:
        """String representation."""
        n_children = len(self._children)
        n_stations = len(self._stations)
        parts = []
        if n_children:
            parts.append(f"{n_children} groups")
        if n_stations:
            parts.append(f"{n_stations} stations")
        return f"<StationNamespace: {', '.join(parts) or 'empty'}>"

    def _ipython_key_completions_(self) -> list[str]:
        """Support bracket completion in IPython."""
        return self.__dir__()


def _fips_to_iso(fips_code: str) -> str:
    """Convert FIPS country code to ISO 3166-1 alpha-2 code.

    Args:
        fips_code: FIPS country code from ISD data

    Returns:
        ISO 3166-1 alpha-2 country code, or original if no mapping exists
    """
    from weathervault._constants import FIPS_TO_ISO

    return FIPS_TO_ISO.get(fips_code, fips_code)


class _StationRegistry(_StationNamespace):
    """Root registry that lazily builds the station hierarchy.

    Stations are organized using ISO 3166-1 alpha-2 country codes:
    - station.<ISO_CODE>.<STATE>.<STATION_NAME> for US stations
    - station.<ISO_CODE>.<STATION_NAME> for other countries

    Examples:
        >>> station.US.NY.LAGUARDIA_AP
        '725030-14732'
        >>> station.DE.BERLIN_TEGEL  # Germany (ISO: DE, FIPS: GM)
        '103820-99999'
    """

    def __init__(self) -> None:
        """Initialize empty registry (loads on first access)."""
        super().__init__()
        self._loaded = False

    def _ensure_loaded(self) -> None:
        """Load station data if not already loaded."""
        if self._loaded:
            return

        df = _get_stations_df()

        # Sort by end_date descending so active stations are processed last
        # (and thus override older stations with the same name)
        df = df.sort("end_date", descending=False, nulls_last=False)

        # Build hierarchy by ISO country code
        country_groups: dict[str, dict[str, dict[str, str] | str]] = {}

        for row in df.iter_rows(named=True):
            station_id = row["id"]
            name = row["name"] or ""
            fips_code = row["country_code"] or "XX"
            # Convert FIPS to ISO for user-facing interface
            country_code = _fips_to_iso(fips_code)
            state = row["state"]

            sanitized_name = _sanitize_name(name)

            if country_code not in country_groups:
                country_groups[country_code] = {}

            # For US stations, add state level
            if country_code == "US" and state:
                if state not in country_groups[country_code]:
                    country_groups[country_code][state] = {}
                state_dict = country_groups[country_code][state]
                if isinstance(state_dict, dict):  # pragma: no branch
                    # Overwrite if same name (newer station replaces older)
                    state_dict[sanitized_name] = station_id
            else:
                # For non-US, put directly under country
                country_groups[country_code][sanitized_name] = station_id

        # Build namespace tree
        for country_code, content in country_groups.items():
            children: dict[str, _StationNamespace] = {}
            stations: dict[str, str] = {}

            for key, value in content.items():
                if isinstance(value, dict):
                    # It's a state with stations
                    children[key] = _StationNamespace(stations=value)
                else:
                    # It's a station directly under country
                    stations[key] = value

            self._children[country_code] = _StationNamespace(stations=stations, children=children)

        self._loaded = True

    def __getattr__(self, name: str) -> _StationNamespace | str:
        """Get a child namespace or station ID, loading data if needed."""
        if name.startswith("_"):
            raise AttributeError(name)
        self._ensure_loaded()
        return super().__getattr__(name)

    def __dir__(self) -> list[str]:
        """Return list of available attributes for autocomplete."""
        self._ensure_loaded()
        return super().__dir__()

    def __repr__(self) -> str:
        """String representation."""
        if not self._loaded:
            return "<StationRegistry: not loaded (access an attribute to load)>"
        return f"<StationRegistry: {len(self._children)} countries>"

    def search(self, pattern: str) -> dict[str, str]:
        """Search for stations matching a pattern.

        Args:
            pattern: Case-insensitive pattern to match against station names

        Returns:
            Dict mapping full paths to station IDs

        Example:
            >>> station.search("laguardia")
            {'US.NY.LAGUARDIA_AP': '725030-14732'}
        """
        self._ensure_loaded()
        results: dict[str, str] = {}
        pattern_upper = pattern.upper()

        for country_code, country_ns in self._children.items():
            # Check stations directly under country
            for name, station_id in country_ns._stations.items():
                if pattern_upper in name.upper():
                    results[f"{country_code}.{name}"] = station_id

            # Check states/children
            for state, state_ns in country_ns._children.items():
                for name, station_id in state_ns._stations.items():
                    if pattern_upper in name.upper():
                        results[f"{country_code}.{state}.{name}"] = station_id

        return results


# Singleton instance
station = _StationRegistry()



"""Main weather data retrieval functions."""

from __future__ import annotations

import contextlib
import importlib.resources
import warnings
from datetime import date
from pathlib import Path
from typing import TYPE_CHECKING

import httpx
import polars as pl

from weathervault._constants import BASE_URL
from weathervault._parsing import _empty_processed_dataframe, parse_isd_data, process_weather_data
from weathervault.stations import get_station_metadata, get_years_for_station

if TYPE_CHECKING:
    pass


def _normalize_temp_unit(temp_unit: str) -> str | None:
    """
    Normalize temperature unit string to standard form.

    Parameters
    ----------
    temp_unit
        Temperature unit string (e.g., `"c"`, `"C"`, `"celsius"`, `"Celsius"`).

    Returns
    -------
    str | None
        Normalized form ("celsius", "fahrenheit", or "kelvin"), or None if invalid.
    """
    unit = temp_unit.lower().strip()
    if unit in ("c", "celsius"):
        return "celsius"
    elif unit in ("f", "fahrenheit"):
        return "fahrenheit"
    elif unit in ("k", "kelvin"):
        return "kelvin"
    return None


def _get_bundled_years(station_id: str) -> set[int]:
    """
    Get years available in bundled sample data for a station.

    Parameters
    ----------
    station_id
        Station ID in USAF-WBAN format.

    Returns
    -------
    set[int]
        Set of years available in bundled data, empty if none.
    """
    try:
        data_files = importlib.resources.files("weathervault.data")
        bundled_years = set()
        for item in data_files.iterdir():
            name = item.name
            if name.startswith(f"{station_id}-") and name.endswith(".gz"):
                # Extract year from filename like "725030-14732-2023.gz"
                year_str = name.replace(f"{station_id}-", "").replace(".gz", "")
                with contextlib.suppress(ValueError):
                    bundled_years.add(int(year_str))
        return bundled_years
    except (ModuleNotFoundError, FileNotFoundError, TypeError):
        return set()


def get_weather_data(
    station_id: str,
    years: int | list[int] | None = None,
    *,
    convert_to_local: bool = True,
    temp_unit: str = "celsius",
    make_hourly: bool = False,
    cache_dir: str | Path | None = None,
    incl_stn_info: bool = False,
    time_as_columns: bool = False,
    quiet: bool = False,
) -> pl.DataFrame:
    """
    Get weather data for a meteorological station.

    Downloads and parses weather observation data from NOAA's Integrated
    Surface Database (ISD) for the specified station and years.

    Parameters
    ----------
    station_id
        Station identifier in USAF-WBAN format (e.g., "725030-14732"). Use `search_stations()` or
        `get_station_metadata()` to find station IDs.
    years
        Year or list of years to retrieve data for. If `None`, retrieves all available years for
        the station.
    convert_to_local
        Whether to convert times from UTC to the station's local time. Default is `True`.
    temp_unit
        Temperature unit for output. Accepts `"c"`/`"celsius"`, `"f"`/`"fahrenheit"`, or
        `"k"`/`"kelvin"` (case-insensitive). Default is `"c"` (Celsius).
    make_hourly
        If `True`, resample data to hourly intervals, taking the first observation of each hour and
        filling missing hours with `None` values. Default is `False`.
    cache_dir
        Directory to cache downloaded data files. Options are: (1) `None` (default) for downloading
        data to memory without saving to disk (will check current working directory for existing
        cached files first), (2) `"."` saves to and loads from current working directory, and (3) a
        path string serving as a custom directory path for caching files. Cached files are named as
        `"{station_id}-{year}.gz"`.
    incl_stn_info
        Whether to include station metadata in the output. When True, adds columns for station name,
        country, state, ICAO code, latitude, longitude, and elevation. Default is `False`.
    time_as_columns
        If `True`, output separate columns for time components (`year`, `month`, `day`, `hour`,
        `min`) instead of a single `time` column. Default is `False`.
    quiet
        If `True`, suppress warning messages about partial data availability (when some requested
        years are not available). Default is `False`.

    Returns
    -------
    pl.DataFrame
        A DataFrame with weather observations containing:

        - `id`: Station identifier
        - `time`: Observation time (local time if `convert_to_local=True`, else UTC); or if
          `time_as_columns=True`: `year`, `month`, `day`, `hour`, `min` columns instead
        - `temp`: Air temperature (Celsius, Fahrenheit, or Kelvin based on `temp_unit=`)
        - `dew_point`: Dew point temperature
        - `rh`: Relative humidity (percentage)
        - `wd`: Wind direction in degrees (`0`-`360`, direction wind is blowing from)
        - `ws`: Wind speed in meters per second
        - `atmos_pres`: Atmospheric pressure at sea level (hectopascals)
        - `ceil_hgt`: Ceiling height in meters (`22000` is taken as unlimited)
        - `visibility`: Visibility in meters (max is `16000`)

        If incl_stn_info=True, also includes:

        - `name`: Station name
        - `country`: Country name
        - `state`: State/region code (if applicable)
        - `icao`: ICAO identifier
        - `lat`: Latitude in degrees
        - `lon`: Longitude in degrees
        - `elev`: Elevation in meters above sea level

    Examples
    --------
    Get weather data for JFK Airport (New York) for 2023:

    ```{python}
    import weathervault as wv

    weather = wv.get_weather_data("744860-94789", years=2023)
    weather.head()
    ```

    Get data for multiple years with Fahrenheit temperatures:

    ```{python}
    weather = wv.get_weather_data(
        "725030-14732",  # LaGuardia Airport
        years=[2022, 2023],
        temp_unit="fahrenheit"
    )

    weather
    ```

    Get hourly data (resampled):

    ```{python}
    weather = wv.get_weather_data(
        "724050-13743",  # Washington Reagan Airport
        years=2023,
        make_hourly=True
    )

    weather
    ```

    Cache data to current directory for offline use:

    ```{python}
    weather = wv.get_weather_data(
        "725030-14732",
        years=2023,
        cache_dir="."  # Saves to current working directory
    )

    # Subsequent calls will automatically use the cached file
    weather = wv.get_weather_data("725030-14732", years=2023)  # Loads from cache

    weather
    ```

    Include station metadata in the output:

    ```{python}
    weather = wv.get_weather_data(
        "725030-14732",
        years=2023,
        incl_stn_info=True
    )

    weather[["name", "country", "lat", "lon"]].head(1)
    ```
    """
    # Normalize and validate temperature unit
    temp_unit_normalized = _normalize_temp_unit(temp_unit)
    if temp_unit_normalized is None:
        raise ValueError("temp_unit must be 'c', 'celsius', 'f', 'fahrenheit', 'k', or 'kelvin'")

    # Get station metadata to find timezone and potentially include in output
    stations = get_station_metadata()
    station_info = stations.filter(pl.col("id") == station_id)

    if station_info.height == 0:
        raise ValueError(
            f"Station '{station_id}' not found in metadata. "
            "Use search_stations() to find valid station IDs."
        )

    tz_name = station_info["tz_name"][0] if "tz_name" in station_info.columns else None

    # Extract station metadata if requested
    stn_metadata = None
    if incl_stn_info:
        stn_metadata = {
            "name": station_info["name"][0],
            "country": station_info["country"][0],
            "state": station_info["state"][0],
            "icao": station_info["icao"][0],
            "lat": station_info["lat"][0],
            "lon": station_info["lon"][0],
            "elev": station_info["elev"][0],
        }

    # Check if we have bundled data for this station (for offline docs)
    bundled_years = _get_bundled_years(station_id)

    # Determine years to fetch
    if years is None:
        # When years=None, we need to know all available years from inventory
        available_years = get_years_for_station(station_id)
        requested_years = available_years
        if not requested_years:
            raise ValueError(f"No data available for station '{station_id}'")
    elif isinstance(years, int):
        requested_years = [years]
    else:
        requested_years = list(years)

    # Start with the requested years as years to fetch
    years_to_fetch = list(requested_years)

    # Check for unavailable years and warn user (skip if all requested years are bundled)
    if years is not None and not set(requested_years).issubset(bundled_years):
        available_years = get_years_for_station(station_id)

        if available_years:
            requested_set = set(requested_years)
            available_set = set(available_years)
            unavailable_years = requested_set - available_set
            fetchable_years = requested_set & available_set

            if unavailable_years:
                unavailable_sorted = sorted(unavailable_years)
                available_range = f"{min(available_years)}-{max(available_years)}"

                if not fetchable_years:
                    # None of the requested years are available
                    raise ValueError(
                        f"No data available for station '{station_id}' for year(s) "
                        f"{', '.join(map(str, unavailable_sorted))}. "
                        f"Available years: {available_range} "
                        f"({len(available_years)} years total). "
                        f"Use get_years_for_station('{station_id}') to see all available years."
                    )
                else:
                    # Some years are available, warn about partial data
                    fetchable_sorted = sorted(fetchable_years)
                    if not quiet:
                        warnings.warn(
                            f"Partial data returned for station '{station_id}'. "
                            f"Requested years {', '.join(map(str, unavailable_sorted))} "
                            f"are not available. "
                            f"Returning data for: {', '.join(map(str, fetchable_sorted))}. "
                            f"Available years for this station: {available_range}.",
                            UserWarning,
                            stacklevel=2,
                        )
                    # Update both requested and fetch lists
                    requested_years = fetchable_sorted
                    years_to_fetch = fetchable_sorted
        elif not bundled_years:
            # Station exists in metadata but has no inventory data and no bundled data
            raise ValueError(
                f"No data inventory found for station '{station_id}'. "
                f"This station may not have any downloadable data files."
            )

    # When converting to local time, we need buffer years to handle timezone offsets
    # at year boundaries. For example, UTC records from Dec 31 23:00 might become
    # Jan 1 08:00 local time (UTC+9), or Jan 1 05:00 UTC might become Dec 31 21:00
    # local time (UTC-8). By fetching adjacent years, we ensure complete coverage.
    if convert_to_local and years_to_fetch:
        buffer_years = set(years_to_fetch)
        for year in requested_years:
            buffer_years.add(year - 1)  # Previous year for positive UTC offsets
            buffer_years.add(year + 1)  # Next year for negative UTC offsets
        years_to_fetch = sorted(buffer_years)

    # Set up cache directory if specified
    # Special case: "." means use current working directory
    if cache_dir == ".":
        cache_path = Path.cwd()
    elif cache_dir:
        cache_path = Path(cache_dir)
        cache_path.mkdir(parents=True, exist_ok=True)
    else:
        cache_path = None

    # Download and parse data for each year
    all_data: list[pl.DataFrame] = []

    for year in years_to_fetch:
        data = _fetch_year_data(station_id, year, cache_path)
        if data is not None:
            df = parse_isd_data(data)
            if df.height > 0:
                processed = process_weather_data(
                    df,
                    tz_name=tz_name,
                    convert_to_local=convert_to_local,
                    temp_unit=temp_unit_normalized,
                    stn_metadata=stn_metadata,
                )
                all_data.append(processed)

    if not all_data:
        return _empty_processed_dataframe(
            incl_stn_info=incl_stn_info, time_as_columns=time_as_columns
        )

    # Combine all years
    result = pl.concat(all_data)

    # Sort by time
    result = result.sort("time")

    # Filter to requested years only (trimming buffer years used for timezone handling)
    # This ensures users get exactly the years they asked for, with complete data
    # at year boundaries after timezone conversion
    result = result.filter(pl.col("time").dt.year().is_in(requested_years))

    # Optionally resample to hourly
    if make_hourly:
        result = _make_hourly(
            result, requested_years, tz_name if convert_to_local else "UTC", incl_stn_info
        )

    # Convert time to separate columns if requested
    if time_as_columns:
        result = _convert_time_to_columns(result, incl_stn_info)

    return result


def _fetch_year_data(
    station_id: str,
    year: int,
    cache_path: Path | None = None,
) -> bytes | None:
    """
    Fetch weather data for a station and year.

    First checks for data in:

    1. Bundled sample data (shipped with package for documentation examples)
    2. The specified cache directory (if provided)
    3. The current working directory

    If not found locally, downloads from NCEI and caches if `cache_path=` is set.

    Parameters
    ----------
    station_id
        Station ID in USAF-WBAN format.
    year
        Year to fetch.
    cache_path
        Optional path to cache directory. If provided, downloaded files will be saved here. If
        `None`, files are not saved after download.

    Returns
    -------
    bytes | None
        Raw gzipped data bytes, or `None` if fetch failed.
    """
    filename = f"{station_id}-{year}.gz"

    # Check for bundled sample data first (used for documentation examples)
    try:
        data_files = importlib.resources.files("weathervault.data")
        bundled_file = data_files.joinpath(filename)
        if bundled_file.is_file():
            return bundled_file.read_bytes()
    except (ModuleNotFoundError, FileNotFoundError, TypeError):
        # No bundled data available, continue to other sources
        pass

    # Check cache directory (if provided)
    if cache_path:
        cached_file = cache_path / filename
        if cached_file.exists():
            return cached_file.read_bytes()

    # Always check current working directory for existing cached files
    # This allows offline use if files were previously downloaded
    cwd_cached_file = Path.cwd() / filename
    if cwd_cached_file.exists():
        return cwd_cached_file.read_bytes()

    # Download from NCEI
    url = f"{BASE_URL}/{year}/{filename}"

    try:
        with httpx.Client(timeout=120.0, follow_redirects=True) as client:
            response = client.get(url)
            response.raise_for_status()
            data = response.content

            # Cache if directory specified
            if cache_path and data:
                cached_file = cache_path / filename
                cached_file.write_bytes(data)

            return data

    except httpx.HTTPStatusError as e:
        if e.response.status_code == 404:
            # Data not available for this year
            return None
        raise
    except httpx.RequestError:
        # Network error
        return None


def _make_hourly(
    df: pl.DataFrame,
    years: list[int],
    tz_name: str | None,
    incl_stn_info: bool = False,
) -> pl.DataFrame:
    """
    Resample data to hourly intervals, filling missing hours.

    Parameters
    ----------
    df
        Weather DataFrame to resample.
    years
        List of years to ensure complete hourly coverage.
    tz_name
        Timezone name for the time column.

    Returns
    -------
    pl.DataFrame
        Resampled hourly DataFrame.
    """
    if df.height == 0:
        return df

    # Get station ID
    station_id = df["id"][0]

    # Truncate time to hour
    df = df.with_columns(pl.col("time").dt.truncate("1h").alias("time"))

    # Build aggregation list
    agg_cols = [
        pl.col("id").first(),
        pl.col("temp").first(),
        pl.col("dew_point").first(),
        pl.col("rh").first(),
        pl.col("wd").first(),
        pl.col("ws").first(),
        pl.col("atmos_pres").first(),
        pl.col("ceil_hgt").first(),
        pl.col("visibility").first(),
    ]

    # Add station metadata columns if present
    if incl_stn_info and "name" in df.columns:
        agg_cols.extend(
            [
                pl.col("name").first(),
                pl.col("country").first(),
                pl.col("state").first(),
                pl.col("icao").first(),
                pl.col("lat").first(),
                pl.col("lon").first(),
                pl.col("elev").first(),
            ]
        )

    # Take first observation of each hour
    df = df.group_by("time").agg(agg_cols)

    # Create complete hourly range for all years
    min_year = min(years)
    max_year = max(years)

    start_time = date(min_year, 1, 1)
    end_time = date(max_year, 12, 31)

    # Generate all hours in the range
    all_hours = pl.datetime_range(
        start=start_time,
        end=end_time,
        interval="1h",
        time_zone=tz_name,
        eager=True,
    )

    hours_df = pl.DataFrame({"time": all_hours})

    # Filter to only the requested years
    hours_df = hours_df.filter(pl.col("time").dt.year().is_in(years))

    # Join with actual data
    result = hours_df.join(df, on="time", how="left")

    # Fill station ID for missing rows
    result = result.with_columns(pl.lit(station_id).alias("id"))

    # Build column order
    columns = [
        "id",
        "time",
        "temp",
        "dew_point",
        "rh",
        "wd",
        "ws",
        "atmos_pres",
        "ceil_hgt",
        "visibility",
    ]

    # Add station metadata columns if present
    if incl_stn_info and "name" in result.columns:
        columns.extend(["name", "country", "state", "icao", "lat", "lon", "elev"])

    # Reorder columns
    result = result.select(columns)

    return result.sort("time")


def _convert_time_to_columns(df: pl.DataFrame, incl_stn_info: bool = False) -> pl.DataFrame:
    """
    Convert time column to separate year, month, day, hour, min columns.

    Parameters
    ----------
    df
        Weather DataFrame with a `time` column.
    incl_stn_info
        Whether station metadata columns are present.

    Returns
    -------
    pl.DataFrame
        DataFrame with time split into separate columns.
    """
    if df.height == 0:
        return _empty_processed_dataframe(incl_stn_info=incl_stn_info, time_as_columns=True)

    # Extract time components
    df = df.with_columns(
        [
            pl.col("time").dt.year().alias("year"),
            pl.col("time").dt.month().alias("month"),
            pl.col("time").dt.day().alias("day"),
            pl.col("time").dt.hour().alias("hour"),
            pl.col("time").dt.minute().alias("min"),
        ]
    )

    # Build column order
    columns = [
        "id",
        "year",
        "month",
        "day",
        "hour",
        "min",
        "temp",
        "dew_point",
        "rh",
        "wd",
        "ws",
        "atmos_pres",
        "ceil_hgt",
        "visibility",
    ]

    # Add station metadata columns if present
    if incl_stn_info and "name" in df.columns:
        columns.extend(["name", "country", "state", "icao", "lat", "lon", "elev"])

    return df.select(columns)



"""Tests for the constants module."""

from weathervault._constants import (
    BASE_URL,
    COUNTRY_CODES,
    MANDATORY_COLUMN_NAMES,
    MANDATORY_COLUMN_WIDTHS,
    MISSING_VALUES,
    SCALE_FACTORS,
)


class TestBaseUrl:
    """Tests for BASE_URL constant."""

    def test_base_url_format(self):
        """Test that BASE_URL is a valid HTTPS URL."""
        assert BASE_URL.startswith("https://")
        assert "ncei.noaa.gov" in BASE_URL

    def test_base_url_path(self):
        """Test that BASE_URL points to the correct data directory."""
        assert "/pub/data/noaa" in BASE_URL

    def test_base_url_no_trailing_slash(self):
        """Test that BASE_URL has no trailing slash for proper path joining."""
        assert not BASE_URL.endswith("/")


class TestColumnWidths:
    """Tests for mandatory column width constants."""

    def test_column_widths_and_names_match(self):
        """Test that column widths and names have the same length."""
        assert len(MANDATORY_COLUMN_WIDTHS) == len(MANDATORY_COLUMN_NAMES)

    def test_column_widths_sum(self):
        """Test that column widths sum to expected mandatory section length."""
        # The mandatory section is 105 characters
        assert sum(MANDATORY_COLUMN_WIDTHS) == 105

    def test_column_widths_positive(self):
        """Test that all column widths are positive integers."""
        for width in MANDATORY_COLUMN_WIDTHS:
            assert isinstance(width, int)
            assert width > 0

    def test_mandatory_column_names_unique(self):
        """Test that column names are unique."""
        assert len(MANDATORY_COLUMN_NAMES) == len(set(MANDATORY_COLUMN_NAMES))

    def test_expected_column_names_present(self):
        """Test that key column names are present."""
        expected = [
            "usaf",
            "wban",
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "wind_direction",
            "wind_speed",
            "temp",
            "dew_point",
            "visibility",
            "ceiling_height",
            "sea_level_pressure",
        ]
        for name in expected:
            assert name in MANDATORY_COLUMN_NAMES


class TestCountryCodes:
    """Tests for COUNTRY_CODES constant."""

    def test_country_codes_not_empty(self):
        """Test that country codes dictionary is populated."""
        assert len(COUNTRY_CODES) > 100  # Should have many countries

    def test_country_codes_format(self):
        """Test that country codes are 2-character strings."""
        for code in COUNTRY_CODES:
            assert len(code) == 2
            assert code.isupper()

    def test_country_names_not_empty(self):
        """Test that country names are non-empty strings."""
        for name in COUNTRY_CODES.values():
            assert isinstance(name, str)
            assert len(name) > 0

    def test_known_country_codes(self):
        """Test some known country code mappings."""
        assert COUNTRY_CODES["US"] == "United States"
        assert COUNTRY_CODES["UK"] == "United Kingdom"
        assert COUNTRY_CODES["CA"] == "Canada"
        assert COUNTRY_CODES["AS"] == "Australia"  # NOTE: FIPS code for Australia is AS
        assert COUNTRY_CODES["AU"] == "Austria"  # NOTE: AU is Austria in FIPS
        assert COUNTRY_CODES["FR"] == "France"
        assert COUNTRY_CODES["GM"] == "Germany"
        assert COUNTRY_CODES["JA"] == "Japan"

    def test_no_duplicate_country_names(self):
        """Test that there are no exact duplicate country names."""
        names = list(COUNTRY_CODES.values())
        # Note: Some countries may have similar names, but exact duplicates indicate errors
        # Allow for some duplicates due to historical/political reasons
        unique_names = set(names)
        # Most names should be unique
        assert len(unique_names) > len(names) * 0.9


class TestMissingValues:
    """Tests for MISSING_VALUES constant."""

    def test_missing_values_keys(self):
        """Test that missing values are defined for expected fields."""
        expected_keys = {
            "wind_direction",
            "wind_speed",
            "ceiling_height",
            "visibility",
            "temp",
            "dew_point",
            "sea_level_pressure",
        }
        assert set(MISSING_VALUES.keys()) == expected_keys

    def test_missing_values_are_integers(self):
        """Test that missing values are integers."""
        for value in MISSING_VALUES.values():
            assert isinstance(value, int)

    def test_missing_values_are_positive(self):
        """Test that missing values are positive (they're typically 9-filled)."""
        for value in MISSING_VALUES.values():
            assert value > 0

    def test_missing_values_pattern(self):
        """Test that missing values follow the 9-filled pattern."""
        # ISD uses 9s for missing values
        for key, value in MISSING_VALUES.items():
            str_value = str(value)
            assert all(c == "9" for c in str_value), f"{key} should be all 9s"


class TestScaleFactors:
    """Tests for SCALE_FACTORS constant."""

    def test_scale_factors_keys(self):
        """Test that scale factors are defined for expected fields."""
        expected_keys = {
            "wind_speed",
            "temp",
            "dew_point",
            "sea_level_pressure",
            "latitude",
            "longitude",
            "elevation",
        }
        assert set(SCALE_FACTORS.keys()) == expected_keys

    def test_scale_factors_are_positive(self):
        """Test that scale factors are positive numbers."""
        for value in SCALE_FACTORS.values():
            assert value > 0

    def test_scale_factors_are_numeric(self):
        """Test that scale factors are integers or floats."""
        for value in SCALE_FACTORS.values():
            assert isinstance(value, (int, float))

    def test_temperature_scale_factor(self):
        """Test that temperature scale factor is 10 (for 1 decimal place)."""
        assert SCALE_FACTORS["temp"] == 10
        assert SCALE_FACTORS["dew_point"] == 10

    def test_wind_speed_scale_factor(self):
        """Test that wind speed scale factor is 10."""
        assert SCALE_FACTORS["wind_speed"] == 10

    def test_coordinate_scale_factor(self):
        """Test that lat/lon scale factors are 1000 (for 3 decimal places)."""
        assert SCALE_FACTORS["latitude"] == 1000
        assert SCALE_FACTORS["longitude"] == 1000
